# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Private helper methods for B |eacute| zier curves.

.. |eacute| unicode:: U+000E9 .. LATIN SMALL LETTER E WITH ACUTE
   :trim:
"""


import functools

import numpy as np
import six

try:
    import scipy.integrate as _scipy_int
except ImportError:  # pragma: NO COVER
    _scipy_int = None

from bezier import _helpers


_MAX_LOCATE_SUBDIVISIONS = 20


def make_subdivision_matrix(degree):
    """Make the matrix used to subdivide a curve.

    Args:
        degree (int): The degree of the curve.

    Returns:
        numpy.ndarray: The matrix used to convert the
           nodes into left and right nodes.
    """
    num_rows = 2 * degree + 1
    result = np.zeros((num_rows, degree + 1))
    result[0, 0] = 1.0
    result[-1, -1] = 1.0
    for row in six.moves.xrange(1, degree + 1):
        half_prev = 0.5 * result[row - 1, :row]
        result[row, :row] = half_prev
        result[row, 1:row + 1] += half_prev
        # Populate the complement row as well.
        complement = num_rows - row - 1
        # NOTE: We "should" reverse the results when using
        #       the complement, but they are symmetric so
        #       that would be a waste.
        result[complement, -(row + 1):] = result[row, :row + 1]
    return result


def evaluate_multi(nodes, degree, s_vals):
    r"""Computes multiple points along a curve.

    Does so by computing the Bernstein basis at each value in ``s_vals``
    rather than using the de Casteljau algorithm.

    Args:
        nodes (numpy.ndarray): The nodes defining a curve.
        degree (int): The degree of the curve (assumed to be one less than
            the number of ``nodes``.
        s_vals (numpy.ndarray): Parameters along the curve (as a
            1D array).

    Returns:
        numpy.ndarray: The evaluated points on the curve as a two dimensional
        NumPy array, with the rows corresponding to each ``s``
        value and the columns to the dimension.
    """
    num_vals, = s_vals.shape

    lambda2 = s_vals[:, np.newaxis]
    lambda1 = 1.0 - lambda2

    weights_next = np.zeros((num_vals, degree + 1))
    weights_curr = np.zeros((num_vals, degree + 1))
    weights_curr[:, 0] = 1.0

    # Increase from degree 0 to ``degree``.
    for curr_deg in six.moves.xrange(degree):
        weights_next[:, :curr_deg + 1] = (
            lambda1 * weights_curr[:, :curr_deg + 1])
        weights_next[:, 1:curr_deg + 2] += (
            lambda2 * weights_curr[:, :curr_deg + 1])
        weights_curr, weights_next = weights_next, weights_curr

    return weights_curr.dot(nodes)


def _vec_size(nodes, degree, s_val):
    r"""Compute :math:`\|B(s)\|_2`.

    Args:
        nodes (numpy.ndarray): The nodes defining a curve.
        degree (int): The degree of the curve (assumed to be one less than
            the number of ``nodes``.
        s_val (float): Parameter to compute :math:`B(s)`.

    Returns:
        float: The norm of :math:`B(s)`.
    """
    result_vec = evaluate_multi(nodes, degree, np.array([s_val]))
    return np.linalg.norm(result_vec, ord=2)


def compute_length(nodes, degree):
    r"""Approximately compute the length of a curve.

    .. _QUADPACK: https://en.wikipedia.org/wiki/QUADPACK

    If ``degree`` is :math:`n`, then the Hodograph curve
    :math:`B'(s)` is degree :math:`d = n - 1`. Using this curve, we
    approximate the integral:

    .. math::

       \ell\left(B\right) =
           \int_0^1 \| B'(s) \|_2 \, ds

    using `QUADPACK`_ (via SciPy).

    Args:
        nodes (numpy.ndarray): The nodes defining a curve.
        degree (int): The degree of the curve (assumed to be one less than
            the number of ``nodes``.

    Returns:
        float: The length of the curve.

    Raises:
        OSError: If SciPy is not installed.
    """
    # NOTE: We somewhat replicate code in ``evaluate_hodograph()``
    #       here. This is so we don't re-compute the nodes for the first
    #       derivative every time it is evaluated.
    first_deriv = degree * (nodes[1:, :] - nodes[:-1, :])
    if degree == 1:
        return np.linalg.norm(first_deriv, ord=2)

    if _scipy_int is None:
        raise OSError('This function requires SciPy for quadrature.')

    size_func = functools.partial(_vec_size, first_deriv, degree - 1)
    length, _ = _scipy_int.quad(size_func, 0.0, 1.0)
    return length


def elevate_nodes(nodes, degree, dimension):
    r"""Degree-elevate a B |eacute| zier curves.

    Does this by converting the current nodes :math:`v_0, \ldots, v_n`
    to new nodes :math:`w_0, \ldots, w_{n + 1}` where

    .. math::

       \begin{align*}
       w_0 &= v_0 \\
       w_j &= \frac{j}{n + 1} v_{j - 1} + \frac{n + 1 - j}{n + 1} v_j \\
       w_{n + 1} &= v_n
       \end{align*}

    Args:
        nodes (numpy.ndarray): The nodes defining a curve.
        degree (int): The degree of the curve (assumed to be one less than
            the number of ``nodes``.
        dimension (int): The dimension of the curve.

    Returns:
        numpy.ndarray: The nodes of the degree-elevated curve.
    """
    new_nodes = np.zeros((degree + 2, dimension))

    multipliers = np.arange(1, degree + 1, dtype=float)[:, np.newaxis]
    denominator = degree + 1.0
    new_nodes[1:-1, :] = (
        multipliers * nodes[:-1, :] +
        (denominator - multipliers) * nodes[1:, :])
    # Hold off on division until the end, to (attempt to) avoid round-off.
    new_nodes /= denominator

    # After setting the internal nodes (which require division), set the
    # boundary nodes.
    new_nodes[0, :] = nodes[0, :]
    new_nodes[-1, :] = nodes[-1, :]

    return new_nodes


def de_casteljau_one_round(nodes, lambda1, lambda2):
    """Perform one round of de Casteljau's algorithm.

    The weights are assumed to sum to one.

    Args:
        nodes (numpy.ndarray): Control points for a curve.
        lambda1 (float): First barycentric weight on interval.
        lambda2 (float): Second barycentric weight on interval.

    Returns:
        numpy.ndarray: The nodes for a "blended" curve one degree
        lower.
    """
    return lambda1 * nodes[:-1, :] + lambda2 * nodes[1:, :]


def specialize_curve(nodes, degree, start, end):
    """Specialize a curve to a re-parameterization

    Does so by taking two points along the number line and then
    reparameterizing the curve onto the interval formed by the
    start and end points.

    .. note::

       This assumes the curve is degree 1 or greater but doesn't check.

    Args:
        nodes (numpy.ndarray): Control points for a curve.
        degree (int): The degree of the curve.
        start (float): The start point of the interval we are specializing to.
        end (float): The end point of the interval we are specializing to.

    Returns:
        numpy.ndarray: The control points for the specialized curve.
    """
    # Uses start-->0, end-->1 to represent the specialization used.
    weights = (
        (1.0 - start, start),
        (1.0 - end, end),
    )
    partial_vals = {
        (0,): de_casteljau_one_round(nodes, *weights[0]),
        (1,): de_casteljau_one_round(nodes, *weights[1]),
    }

    for _ in six.moves.xrange(degree - 1, 0, -1):
        new_partial = {}
        for key, sub_nodes in six.iteritems(partial_vals):
            # Our keys are ascending so we increment from the last value.
            for next_id in six.moves.xrange(key[-1], 1 + 1):
                new_key = key + (next_id,)
                new_partial[new_key] = de_casteljau_one_round(
                    sub_nodes, *weights[next_id])

        partial_vals = new_partial

    result = np.empty(nodes.shape)
    for index in six.moves.xrange(degree + 1):
        key = (0,) * (degree - index) + (1,) * index
        result[index, :] = partial_vals[key]
    return result


def evaluate_hodograph(nodes, degree, s):
    r"""Evaluate the Hodograph curve at a point :math:`s`.

    The Hodograph (first derivative) of a B |eacute| zier curve
    degree :math:`d = n - 1` and is given by

    .. math::

       B'(s) = n \sum_{j = 0}^{d} \binom{d}{j} s^j
       (1 - s)^{d - j} \cdot \Delta v_j

    where each forward difference is given by
    :math:`\Delta v_j = v_{j + 1} - v_j`.

    Args:
        nodes (numpy.ndarray): The nodes of a curve.
        degree (int): The degree of the curve (assumed to be one less than
            the number of ``nodes``.
        s (float): A parameter along the curve at which the Hodograph
            is to be evaluated.

    Returns:
        numpy.ndarray: The point on the Hodograph curve (as a one
        dimensional NumPy array).
    """
    first_deriv = nodes[1:, :] - nodes[:-1, :]
    # NOTE: Taking the derivative drops the degree by 1.
    return degree * evaluate_multi(
        first_deriv, degree - 1, np.array([s])).flatten()


def get_curvature(nodes, degree, tangent_vec, s):
    r"""Compute the signed curvature of a curve at :math:`s`.

    Computed via

    .. math::

       \frac{B'(s) \times B''(s)}{\left\lVert B'(s) \right\rVert_2^3}

    .. image:: images/get_curvature.png
       :align: center

    .. testsetup:: get-curvature

       import numpy as np
       import bezier
       from bezier._curve_helpers import evaluate_hodograph
       from bezier._curve_helpers import get_curvature

       def hodograph(nodes, s):
           degree = nodes.shape[0] - 1
           return evaluate_hodograph(nodes, degree, s)

    .. doctest:: get-curvature
       :options: +NORMALIZE_WHITESPACE

       >>> nodes = np.array([
       ...     [1.0 ,  0.0],
       ...     [0.75,  2.0],
       ...     [0.5 , -2.0],
       ...     [0.25,  2.0],
       ...     [0.0 ,  0.0],
       ... ])
       >>> s = 0.5
       >>> tangent_vec = hodograph(nodes, s)
       >>> tangent_vec
       array([-1., 0.])
       >>> curvature = get_curvature(nodes, 4, tangent_vec, s)
       >>> curvature
       -12.0

    .. testcleanup:: get-curvature

       import make_images
       make_images.get_curvature(nodes, s, tangent_vec, curvature)

    Args:
        nodes (numpy.ndarray): The nodes of a curve.
        degree (int): The degree of the curve.
        tangent_vec (numpy.ndarray): The already computed value of
            :math:`B'(s)`
        s (float): The parameter value along the curve.

    Returns:
        float: The signed curvature.
    """
    if degree == 1:
        return 0.0

    # NOTE: We somewhat replicate code in ``evaluate_hodograph()``
    #       here. It may be worthwhile to implement ``Curve.hodograph()``
    #       and ``Curve.concavity()`` to avoid re-computing the
    #       first and second node differences.
    first_deriv = nodes[1:, :] - nodes[:-1, :]
    second_deriv = first_deriv[1:, :] - first_deriv[:-1, :]
    concavity = degree * (degree - 1) * evaluate_multi(
        second_deriv, degree - 2, np.array([s]))
    # NOTE: This assumes ``tangent_vec`` was ``flatten()``-ed, but
    #       we intentionally don't flatten ``concavity``.
    curvature = _helpers.cross_product(
        np.array([tangent_vec]), concavity)
    curvature /= np.linalg.norm(tangent_vec, ord=2)**3
    return curvature


def newton_refine(curve, point, s):
    r"""Refine a solution to :math:`B(s) = p` using Newton's method.

    Computes updates via

    .. math::

       \mathbf{0} \approx
           \left(B\left(s_{\ast}\right) - p\right) +
           B'\left(s_{\ast}\right) \Delta s

    For example, consider the curve

    .. math::

       B(s) =
           \left[\begin{array}{c} 0 \\ 0 \end{array}\right] (1 - s)^2 +
           \left[\begin{array}{c} 1 \\ 2 \end{array}\right]
               2 (1 - s) s +
           \left[\begin{array}{c} 3 \\ 1 \end{array}\right] s^2

    and the point :math:`B\left(\frac{1}{4}\right) =
    \frac{1}{16} \left[\begin{array}{c} 9 \\ 13 \end{array}\right]`.

    Starting from the **wrong** point :math:`s = \frac{3}{4}`, we have

    .. math::

       \begin{align*}
       p - B\left(\frac{1}{2}\right) &= -\frac{1}{2}
           \left[\begin{array}{c} 3 \\ 1 \end{array}\right] \\
       B'\left(\frac{1}{2}\right) &= \frac{1}{2}
           \left[\begin{array}{c} 7 \\ -1 \end{array}\right] \\
       \Longrightarrow \frac{1}{4} \left[\begin{array}{c c}
           7 & -1 \end{array}\right] \left[\begin{array}{c}
           7 \\ -1 \end{array}\right] \Delta s &= -\frac{1}{4}
           \left[\begin{array}{c c} 7 & -1 \end{array}\right]
           \left[\begin{array}{c} 3 \\ 1 \end{array}\right] \\
       \Longrightarrow \Delta s &= -\frac{2}{5}
       \end{align*}

    .. image:: images/newton_refine_curve.png
       :align: center

    .. testsetup:: newton-refine-curve, newton-refine-curve-cusp

       import numpy as np
       import bezier
       from bezier._curve_helpers import newton_refine

    .. doctest:: newton-refine-curve
       :options: +NORMALIZE_WHITESPACE

       >>> curve = bezier.Curve(np.array([
       ...     [0.0, 0.0],
       ...     [1.0, 2.0],
       ...     [3.0, 1.0],
       ... ]))
       >>> point = curve.evaluate_multi(np.array([0.25]))
       >>> point
       array([[ 0.5625, 0.8125]])
       >>> s = 0.75
       >>> new_s = newton_refine(curve, point, s)
       >>> 5 * (new_s - s)
       -2.0

    .. testcleanup:: newton-refine-curve

       import make_images
       make_images.newton_refine_curve(curve, point, s, new_s)

    On curves that are not "valid" (i.e. :math:`B(s)` is not
    injective with non-zero gradient), Newton's method may
    break down and converge linearly:

    .. image:: images/newton_refine_curve_cusp.png
       :align: center

    .. doctest:: newton-refine-curve-cusp
       :options: +NORMALIZE_WHITESPACE

       >>> curve = bezier.Curve(np.array([
       ...     [ 6.0, -3.0],
       ...     [-2.0,  3.0],
       ...     [-2.0, -3.0],
       ...     [ 6.0,  3.0],
       ... ]))
       >>> expected = 0.5
       >>> point = curve.evaluate_multi(np.array([expected]))
       >>> point
       array([[ 0., 0.]])
       >>> s_vals = [0.625, None, None, None, None, None]
       >>> np.log2(abs(expected - s_vals[0]))
       -3.0
       >>> s_vals[1] = newton_refine(curve, point, s_vals[0])
       >>> np.log2(abs(expected - s_vals[1]))
       -3.983...
       >>> s_vals[2] = newton_refine(curve, point, s_vals[1])
       >>> np.log2(abs(expected - s_vals[2]))
       -4.979...
       >>> s_vals[3] = newton_refine(curve, point, s_vals[2])
       >>> np.log2(abs(expected - s_vals[3]))
       -5.978...
       >>> s_vals[4] = newton_refine(curve, point, s_vals[3])
       >>> np.log2(abs(expected - s_vals[4]))
       -6.978...
       >>> s_vals[5] = newton_refine(curve, point, s_vals[4])
       >>> np.log2(abs(expected - s_vals[5]))
       -7.978...

    .. testcleanup:: newton-refine-curve-cusp

       import make_images
       make_images.newton_refine_curve_cusp(curve, s_vals)

    Due to round-off, the Newton process terminates with an error that is not
    close to machine precision :math:`\varepsilon` when :math:`\Delta s = 0`.

    .. testsetup:: newton-refine-curve-cusp-continued

       import numpy as np
       import bezier
       from bezier._curve_helpers import newton_refine

       curve = bezier.Curve(np.array([
           [6.0, -3.0],
           [-2.0, 3.0],
           [-2.0, -3.0],
           [6.0, 3.0],
       ]))
       point = curve.evaluate_multi(np.array([0.5]))

    .. doctest:: newton-refine-curve-cusp

       >>> s_vals = [0.625]
       >>> new_s = newton_refine(curve, point, s_vals[-1])
       >>> while new_s not in s_vals:
       ...     s_vals.append(new_s)
       ...     new_s = newton_refine(curve, point, s_vals[-1])
       ...
       >>> len(s_vals)
       27
       >>> terminal_s = s_vals[-1]
       >>> terminal_s == newton_refine(curve, point, terminal_s)
       True
       >>> np.log2(terminal_s - 0.5)
       -29.580...

    Due to round-off near the cusp, the final error resembles
    :math:`\sqrt{\varepsilon}` rather than machine precision
    as expected.

    Args:
        curve (.Curve): The curve to refine a point on.
        point (numpy.ndarray): A point on ``curve``.
        s (float): An "almost" solution to :math:`B(s) = p`.

    Returns:
        float: The updated value :math:`s + \Delta s`.
    """
    nodes = curve._nodes  # pylint: disable=protected-access
    pt_delta = point.flatten() - curve.evaluate(s)
    derivative = evaluate_hodograph(nodes, curve.degree, s)
    delta_s = np.dot(pt_delta, derivative) / np.dot(derivative, derivative)
    return s + delta_s


def locate_point(curve, point):
    r"""Locate a point on a curve.

    Does so by recursively subdividing the curve and rejecting
    sub-curves with bounding boxes that don't contain the point.
    After the sub-curves are sufficiently small, uses Newton's
    method to zoom in on the parameter value.

    .. note::

       This assumes, but does not check, that ``point`` is ``1xD``,
       where ``D`` is the dimension that ``curve`` is in.

    Args:
        curve (.Curve): A B |eacute| zier curve.
        point (numpy.ndarray): The point to locate.

    Returns:
        Optional[float]: The parameter value (:math:`s`) corresponding
        to ``point`` or :data:`None` if the point is not on the ``curve``.
    """
    point = point.flatten()
    candidates = [curve]
    for _ in six.moves.xrange(_MAX_LOCATE_SUBDIVISIONS + 1):
        next_candidates = []
        for candidate in candidates:
            nodes = candidate._nodes  # pylint: disable=protected-access
            if _helpers.contains_nd(nodes, point):
                next_candidates.extend(candidate.subdivide())

        candidates = next_candidates

    if not candidates:
        return None

    s_approx = np.mean(
        [(candidate.start, candidate.end) for candidate in candidates])
    return newton_refine(curve, point, s_approx)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Curved polygon and associated helpers.

A curved polygon (in :math:`\mathbf{R}^2`) is defined by the
collection of B |eacute| zier curves that determine the
boundary.

.. |eacute| unicode:: U+000E9 .. LATIN SMALL LETTER E WITH ACUTE
   :trim:

.. testsetup:: *

   import numpy as np
   import bezier
"""


from matplotlib import patches
from matplotlib import path as _path_mod
import matplotlib.pyplot as plt
import numpy as np
import six

from bezier import _helpers


class CurvedPolygon(object):
    """Represents an object defined by its curved boundary.

    The boundary is a piecewise defined collection of
    B |eacute| zier curves.

    .. note::

       The direction of the nodes in each :class:`.Curve`
       on the boundary is important: we check that one curve
       begins where the last one ended.

    .. image:: ../images/curved_polygon_constructor1.png
       :align: center

    .. doctest:: curved-polygon-constructor

       >>> import bezier
       >>> edge0 = bezier.Curve(np.array([
       ...     [0.0,  0.0],
       ...     [1.0, -1.0],
       ...     [2.0,  0.0],
       ... ]))
       >>> edge1 = bezier.Curve(np.array([
       ...     [2.0, 0.0],
       ...     [2.0, 1.0],
       ... ]))
       >>> edge2 = bezier.Curve(np.array([
       ...     [2.0, 1.0],
       ...     [1.0, 2.0],
       ...     [0.0, 1.0],
       ... ]))
       >>> edge3 = bezier.Curve(np.array([
       ...     [0.0, 1.0],
       ...     [0.0, 0.0],
       ... ]))
       >>> curved_poly = bezier.CurvedPolygon(
       ...     edge0, edge1, edge2, edge3)
       >>> curved_poly
       <CurvedPolygon (num_sides=4)>

    .. testcleanup:: curved-polygon-constructor

       import make_images
       make_images.curved_polygon_constructor1(curved_poly)

    Though the endpoints of edge pair of edges are verified to match,
    the curved polygon as a whole is not verified, so creating
    a curved polygon with self-intersections is possible:

    .. image:: ../images/curved_polygon_constructor2.png
       :align: center

    .. doctest:: curved-polygon-constructor-invalid

       >>> edge0 = bezier.Curve(np.array([
       ...     [0.0, 0.0],
       ...     [1.0, 0.0],
       ... ]))
       >>> edge1 = bezier.Curve(np.array([
       ...     [1.0 , 0.0],
       ...     [1.25, 0.5],
       ...     [1.0 , 1.0],
       ... ]))
       >>> edge2 = bezier.Curve(np.array([
       ...     [1.0, 1.0],
       ...     [2.0, 1.0],
       ... ]))
       >>> edge3 = bezier.Curve(np.array([
       ...     [2.0, 1.0 ],
       ...     [1.0, 0.75],
       ...     [0.0, 0.0 ],
       ... ]))
       >>> curved_poly = bezier.CurvedPolygon(
       ...     edge0, edge1, edge2, edge3)
       >>> curved_poly
       <CurvedPolygon (num_sides=4)>

    .. testcleanup:: curved-polygon-constructor-invalid

       import make_images
       make_images.curved_polygon_constructor2(curved_poly)

    Args:
        edges (Tuple[~bezier.curve.Curve, ...]): The boundary edges
            of the curved polygon.
    """

    def __init__(self, *edges):
        self._edges = edges
        self._num_sides = len(edges)
        self._verify()

    @staticmethod
    def _verify_pair(prev, curr):
        """Verify a pair of sides share an endpoint.

        .. note::

           This currently checks that edge endpoints match **exactly**
           but allowing some roundoff may be desired.

        Args:
            prev (.Curve): "Previous" curve at piecewise junction.
            curr (.Curve): "Next" curve at piecewise junction.

        Raises:
            ValueError: If the previous side is not in 2D.
            ValueError: If consecutive sides don't share an endpoint.
        """
        if prev.dimension != 2:
            raise ValueError('Curve not in R^2', prev)

        # pylint: disable=protected-access
        prev_nodes = prev._nodes
        curr_nodes = curr._nodes
        # pylint: enable=protected-access
        if not _helpers.vector_close(prev_nodes[-1, :], curr_nodes[0, :]):
            raise ValueError(
                'Not sufficiently close',
                'Consecutive sides do not have common endpoint',
                prev, curr)

    def _verify(self):
        """Verify that the edges define a curved polygon.

        This may not be entirely comprehensive, e.g. won't check
        self-intersection of the defined polygon.

        .. note::

           This currently checks that edge endpoints match **exactly**
           but allowing some roundoff may be desired.

        Raises:
            ValueError: If there are fewer than two sides.
            ValueError: If one of the sides is not in 2D.
            ValueError: If consecutive sides don't share an endpoint.
        """
        if self.num_sides < 2:
            raise ValueError('At least two sides required.')

        for prev, curr in six.moves.zip(self._edges, self._edges[1:]):
            self._verify_pair(prev, curr)

        # Now we check that the final edge wraps around.
        prev = self._edges[-1]
        curr = self._edges[0]
        self._verify_pair(prev, curr)

    @property
    def num_sides(self):
        """int: The number of sides in the current polygon."""
        return self._num_sides

    def __repr__(self):
        """Representation of current object.

        Returns:
            str: Object representation.
        """
        return '<{} (num_sides={:d})>'.format(
            self.__class__.__name__, self.num_sides)

    def plot(self, pts_per_edge, color=None, ax=None, show=False):
        """Plot the current curved polygon.

        Args:
            pts_per_edge (int): Number of points to plot per curved edge.
            color (Optional[Tuple[float, float, float]]): Color as RGB profile.
            ax (Optional[matplotlib.artist.Artist]): matplotlib axis object
                to add plot to.
            show (Optional[bool]): Flag indicating if the plot should be
                shown.

        Returns:
            matplotlib.artist.Artist: The axis containing the plot. This
            may be a newly created axis.
        """
        if ax is None:
            fig = plt.figure()
            ax = fig.gca()

        s_vals = np.linspace(0.0, 1.0, pts_per_edge)

        # Evaluate points on each edge.
        all_points = []
        for edge in self._edges:
            points = edge.evaluate_multi(s_vals)
            line, = ax.plot(points[:, 0], points[:, 1], color=color)
            color = line.get_color()
            # Since the edges overlap, we leave out the first point in each.
            all_points.append(points[1:, :])

        polygon = np.vstack(all_points)
        path = _path_mod.Path(polygon)
        patch = patches.PathPatch(
            path, facecolor=color, alpha=0.6)
        ax.add_patch(patch)

        if show:
            plt.show()

        return ax



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Private helper methods for :mod:`bezier.surface`.

.. |eacute| unicode:: U+000E9 .. LATIN SMALL LETTER E WITH ACUTE
   :trim:
"""


import collections
import enum
import functools
import operator

import numpy as np
import six

from bezier import _curve_helpers
from bezier import _helpers
from bezier import _intersection_helpers
from bezier import curved_polygon


_MAX_POLY_SUBDIVISIONS = 5
_MAX_LOCATE_SUBDIVISIONS = 20
_LOCATE_EPS = 2.0**(-47)
_SIGN = np.sign  # pylint: disable=no-member
_SAME_CURVATURE = 'Tangent curves have same curvature.'
_BAD_TANGENT = (
    'Curves moving in opposite direction but define '
    'overlapping arcs.')
_WRONG_CURVE = 'Start and end node not defined on same curve'
# NOTE: The ``SUBDIVIDE`` matrices are public since used in
#       the ``surface`` module.
LINEAR_SUBDIVIDE = np.array([
    [2, 0, 0],
    [1, 1, 0],
    [0, 2, 0],
    [1, 0, 1],
    [0, 1, 1],
    [0, 0, 2],
], dtype=float) / 2.0
QUADRATIC_SUBDIVIDE = np.array([
    [4, 0, 0, 0, 0, 0],
    [2, 2, 0, 0, 0, 0],
    [1, 2, 1, 0, 0, 0],
    [0, 2, 2, 0, 0, 0],
    [0, 0, 4, 0, 0, 0],
    [2, 0, 0, 2, 0, 0],
    [1, 1, 0, 1, 1, 0],
    [0, 1, 1, 1, 1, 0],
    [0, 0, 2, 0, 2, 0],
    [1, 0, 0, 2, 0, 1],
    [0, 1, 0, 1, 1, 1],
    [0, 0, 1, 0, 2, 1],
    [0, 0, 0, 2, 0, 2],
    [0, 0, 0, 0, 2, 2],
    [0, 0, 0, 0, 0, 4],
], dtype=float) / 4.0
CUBIC_SUBDIVIDE = np.array([
    [8, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    [4, 4, 0, 0, 0, 0, 0, 0, 0, 0],
    [2, 4, 2, 0, 0, 0, 0, 0, 0, 0],
    [1, 3, 3, 1, 0, 0, 0, 0, 0, 0],
    [0, 2, 4, 2, 0, 0, 0, 0, 0, 0],
    [0, 0, 4, 4, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 8, 0, 0, 0, 0, 0, 0],
    [4, 0, 0, 0, 4, 0, 0, 0, 0, 0],
    [2, 2, 0, 0, 2, 2, 0, 0, 0, 0],
    [1, 2, 1, 0, 1, 2, 1, 0, 0, 0],
    [0, 1, 2, 1, 1, 2, 1, 0, 0, 0],
    [0, 0, 2, 2, 0, 2, 2, 0, 0, 0],
    [0, 0, 0, 4, 0, 0, 4, 0, 0, 0],
    [2, 0, 0, 0, 4, 0, 0, 2, 0, 0],
    [1, 1, 0, 0, 2, 2, 0, 1, 1, 0],
    [0, 1, 1, 0, 1, 2, 1, 1, 1, 0],
    [0, 0, 1, 1, 0, 2, 2, 1, 1, 0],
    [0, 0, 0, 2, 0, 0, 4, 0, 2, 0],
    [1, 0, 0, 0, 3, 0, 0, 3, 0, 1],
    [0, 1, 0, 0, 1, 2, 0, 2, 1, 1],
    [0, 0, 1, 0, 0, 2, 1, 1, 2, 1],
    [0, 0, 0, 1, 0, 0, 3, 0, 3, 1],
    [0, 0, 0, 0, 2, 0, 0, 4, 0, 2],
    [0, 0, 0, 0, 0, 2, 0, 2, 2, 2],
    [0, 0, 0, 0, 0, 0, 2, 0, 4, 2],
    [0, 0, 0, 0, 0, 0, 0, 4, 0, 4],
    [0, 0, 0, 0, 0, 0, 0, 0, 4, 4],
    [0, 0, 0, 0, 0, 0, 0, 0, 0, 8],
], dtype=float) / 8.0
QUARTIC_SUBDIVIDE = np.array([
    [16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    [8, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    [4, 8, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    [2, 6, 6, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    [1, 4, 6, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    [0, 2, 6, 6, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    [0, 0, 4, 8, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 8, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    [8, 0, 0, 0, 0, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    [4, 4, 0, 0, 0, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0],
    [2, 4, 2, 0, 0, 2, 4, 2, 0, 0, 0, 0, 0, 0, 0],
    [1, 3, 3, 1, 0, 1, 3, 3, 1, 0, 0, 0, 0, 0, 0],
    [0, 1, 3, 3, 1, 1, 3, 3, 1, 0, 0, 0, 0, 0, 0],
    [0, 0, 2, 4, 2, 0, 2, 4, 2, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 4, 4, 0, 0, 4, 4, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 8, 0, 0, 0, 8, 0, 0, 0, 0, 0, 0],
    [4, 0, 0, 0, 0, 8, 0, 0, 0, 4, 0, 0, 0, 0, 0],
    [2, 2, 0, 0, 0, 4, 4, 0, 0, 2, 2, 0, 0, 0, 0],
    [1, 2, 1, 0, 0, 2, 4, 2, 0, 1, 2, 1, 0, 0, 0],
    [0, 1, 2, 1, 0, 1, 3, 3, 1, 1, 2, 1, 0, 0, 0],
    [0, 0, 1, 2, 1, 0, 2, 4, 2, 1, 2, 1, 0, 0, 0],
    [0, 0, 0, 2, 2, 0, 0, 4, 4, 0, 2, 2, 0, 0, 0],
    [0, 0, 0, 0, 4, 0, 0, 0, 8, 0, 0, 4, 0, 0, 0],
    [2, 0, 0, 0, 0, 6, 0, 0, 0, 6, 0, 0, 2, 0, 0],
    [1, 1, 0, 0, 0, 3, 3, 0, 0, 3, 3, 0, 1, 1, 0],
    [0, 1, 1, 0, 0, 1, 3, 2, 0, 2, 3, 1, 1, 1, 0],
    [0, 0, 1, 1, 0, 0, 2, 3, 1, 1, 3, 2, 1, 1, 0],
    [0, 0, 0, 1, 1, 0, 0, 3, 3, 0, 3, 3, 1, 1, 0],
    [0, 0, 0, 0, 2, 0, 0, 0, 6, 0, 0, 6, 0, 2, 0],
    [1, 0, 0, 0, 0, 4, 0, 0, 0, 6, 0, 0, 4, 0, 1],
    [0, 1, 0, 0, 0, 1, 3, 0, 0, 3, 3, 0, 3, 1, 1],
    [0, 0, 1, 0, 0, 0, 2, 2, 0, 1, 4, 1, 2, 2, 1],
    [0, 0, 0, 1, 0, 0, 0, 3, 1, 0, 3, 3, 1, 3, 1],
    [0, 0, 0, 0, 1, 0, 0, 0, 4, 0, 0, 6, 0, 4, 1],
    [0, 0, 0, 0, 0, 2, 0, 0, 0, 6, 0, 0, 6, 0, 2],
    [0, 0, 0, 0, 0, 0, 2, 0, 0, 2, 4, 0, 4, 2, 2],
    [0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 4, 2, 2, 4, 2],
    [0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 6, 0, 6, 2],
    [0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 8, 0, 4],
    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 4, 4, 4],
    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 8, 4],
    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8, 0, 8],
    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8, 8],
    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16],
], dtype=float) / 16.0
# The Jacobian of a quadratric (in any dimension) as given by
# dB/ds = [-2L1, 2(L1 - L2), 2L2, -2L3, 2L3, 0] * nodes
# dB/dt = [-2L1, -2L2, 0, 2(L1 - L3), 2L2, 2L3] * nodes
# We evaluate this at each of the 6 points in the quadratic
# triangle and then stack them (2 rows * 6 = 12 rows)
# pylint: disable=bad-whitespace
_QUADRATIC_JACOBIAN_HELPER = np.array([
    [-2,  2, 0,  0, 0, 0],
    [-2,  0, 0,  2, 0, 0],
    [-1,  0, 1,  0, 0, 0],
    [-1, -1, 0,  1, 1, 0],
    [ 0, -2, 2,  0, 0, 0],  # noqa: E201
    [ 0, -2, 0,  0, 2, 0],  # noqa: E201
    [-1,  1, 0, -1, 1, 0],
    [-1,  0, 0,  0, 0, 1],
    [ 0, -1, 1, -1, 1, 0],  # noqa: E201
    [ 0, -1, 0, -1, 1, 1],  # noqa: E201
    [ 0,  0, 0, -2, 2, 0],  # noqa: E201
    [ 0,  0, 0, -2, 0, 2],  # noqa: E201
], dtype=float)
_QUADRATIC_TO_BERNSTEIN = np.array([
    [ 2, 0,  0, 0, 0,  0],  # noqa: E201
    [-1, 4, -1, 0, 0,  0],
    [ 0, 0,  2, 0, 0,  0],  # noqa: E201
    [-1, 0,  0, 4, 0, -1],
    [ 0, 0, -1, 0, 4, -1],  # noqa: E201
    [ 0, 0,  0, 0, 0,  2],  # noqa: E201
], dtype=float) / 2.0
# pylint: enable=bad-whitespace
# The Jacobian of a cubic (in any dimension) as given by
# dB/ds = [-3 L1^2, 3 L1(L1 - 2 L2), 3 L2(2 L1 - L2), 3 L2^2, -6 L1 L3,
#          6 L3(L1 - L2), 6 L2 L3, -3 L3^2, 3 L3^2, 0] * nodes
# dB/dt = [-3 L1^2, -6 L1 L2, -3 L2^2, 0, 3 L1(L1 - 2 L3), 6 L2 (L1 - L3),
#          3 L2^2, 3 L3(2 L1 - L3), 6 L2 L3, 3 L3^2] * nodes
# We evaluate this at each of the 15 points in the quartic
# triangle and then stack them (2 rows * 15 = 30 rows)
# pylint: disable=bad-whitespace
_CUBIC_JACOBIAN_HELPER = np.array([
    [-48,  48,   0,  0,   0,   0,  0,   0,  0,  0],
    [-48,   0,   0,  0,  48,   0,  0,   0,  0,  0],
    [-27,   9,  15,  3,   0,   0,  0,   0,  0,  0],
    [-27, -18,  -3,  0,  27,  18,  3,   0,  0,  0],
    [-12, -12,  12, 12,   0,   0,  0,   0,  0,  0],
    [-12, -24, -12,  0,  12,  24, 12,   0,  0,  0],
    [ -3, -15,  -9, 27,   0,   0,  0,   0,  0,  0],  # noqa: E201
    [ -3, -18, -27,  0,   3,  18, 27,   0,  0,  0],  # noqa: E201
    [  0,   0, -48, 48,   0,   0,  0,   0,  0,  0],  # noqa: E201
    [  0,   0, -48,  0,   0,   0, 48,   0,  0,  0],  # noqa: E201
    [-27,  27,   0,  0, -18,  18,  0,  -3,  3,  0],
    [-27,   0,   0,  0,   9,   0,  0,  15,  0,  3],
    [-12,   0,   9,  3, -12,   6,  6,  -3,  3,  0],
    [-12, -12,  -3,  0,   0,   6,  3,   9,  6,  3],
    [ -3,  -9,   0, 12,  -6,  -6, 12,  -3,  3,  0],  # noqa: E201
    [ -3, -12, -12,  0,  -3,   0, 12,   3, 12,  3],  # noqa: E201
    [  0,   0, -27, 27,   0, -18, 18,  -3,  3,  0],  # noqa: E201
    [  0,   0, -27,  0,   0, -18, 27,  -3, 18,  3],  # noqa: E201
    [-12,  12,   0,  0, -24,  24,  0, -12, 12,  0],
    [-12,   0,   0,  0, -12,   0,  0,  12,  0, 12],
    [ -3,  -3,   3,  3, -12,   0, 12, -12, 12,  0],  # noqa: E201
    [ -3,  -6,  -3,  0,  -9,  -6,  3,   0, 12, 12],  # noqa: E201
    [  0,   0, -12, 12,   0, -24, 24, -12, 12,  0],  # noqa: E201
    [  0,   0, -12,  0,   0, -24, 12, -12, 24, 12],  # noqa: E201
    [ -3,   3,   0,  0, -18,  18,  0, -27, 27,  0],  # noqa: E201
    [ -3,   0,   0,  0, -15,   0,  0,  -9,  0, 27],  # noqa: E201
    [  0,   0,  -3,  3,   0, -18, 18, -27, 27,  0],  # noqa: E201
    [  0,   0,  -3,  0,   0, -18,  3, -27, 18, 27],  # noqa: E201
    [  0,   0,   0,  0,   0,   0,  0, -48, 48,  0],  # noqa: E201
    [  0,   0,   0,  0,   0,   0,  0, -48,  0, 48],  # noqa: E201
], dtype=float) / 16.0
# pylint: enable=bad-whitespace
_QUARTIC_TO_BERNSTEIN = np.array([
    [36, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    [-39, 144, -108, 48, -9, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    [26, -128, 240, -128, 26, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    [-9, 48, -108, 144, -39, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 36, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    [-39, 0, 0, 0, 0, 144, 0, 0, 0, -108, 0, 0, 48, 0, -9],
    [26, -64, -24, 32, -9, -64, 288, -96, 16, -24, -96, 12, 32, 16, -9],
    [-9, 32, -24, -64, 26, 16, -96, 288, -64, 12, -96, -24, 16, 32, -9],
    [0, 0, 0, 0, -39, 0, 0, 0, 144, 0, 0, -108, 0, 48, -9],
    [26, 0, 0, 0, 0, -128, 0, 0, 0, 240, 0, 0, -128, 0, 26],
    [-9, 16, 12, 16, -9, 32, -96, -96, 32, -24, 288, -24, -64, -64, 26],
    [0, 0, 0, 0, 26, 0, 0, 0, -128, 0, 0, 240, 0, -128, 26],
    [-9, 0, 0, 0, 0, 48, 0, 0, 0, -108, 0, 0, 144, 0, -39],
    [0, 0, 0, 0, -9, 0, 0, 0, 48, 0, 0, -108, 0, 144, -39],
    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 36],
], dtype=float)
# NOTE: We avoid round-off until after ``_QUARTIC_TO_BERNSTEIN``
#       has been applied.
_QUARTIC_BERNSTEIN_FACTOR = 36.0


def polynomial_sign(poly_surface):
    r"""Determine the "sign" of a polynomial on the reference triangle.

    Checks if a polynomial :math:`p(s, t)` is positive, negative
    or mixed sign on the reference triangle.

    Does this by utilizing the B |eacute| zier form of :math:`p`: it is a
    convex combination of the Bernstein basis (real numbers) hence
    if the Bernstein basis is all positive, the polynomial must be.

    If the values are mixed, then we can recursively subdivide
    until we are in a region where the coefficients are all one
    sign.

    Args:
        poly_surface (Surface): A polynomial on the reference triangle
            specified as a surface.

    Returns:
        int: The sign of the polynomial. Will be one of ``-1``, ``1``
        or ``0``. A value of ``0`` indicates a mixed sign or the
        zero polynomial.

    Raises:
        ValueError: If no conclusion is reached after the maximum
            number of subdivisions.
    """
    sub_polys = [poly_surface]
    signs = set()
    for _ in six.moves.xrange(_MAX_POLY_SUBDIVISIONS):
        undecided = []
        for poly in sub_polys:
            # Avoid an unnecessarily copying the nodes.
            # pylint: disable=protected-access
            nodes = poly._nodes
            # pylint: enable=protected-access
            if np.all(nodes == 0.0):
                signs.add(0)
            elif np.all(nodes > 0.0):
                signs.add(1)
            elif np.all(nodes < 0.0):
                signs.add(-1)
            else:
                undecided.append(poly)

            if len(signs) > 1:
                return 0

        sub_polys = functools.reduce(
            operator.add, [poly.subdivide() for poly in undecided], ())
        if not sub_polys:
            break

    if len(sub_polys) == 0:
        # NOTE: We are guaranteed that ``len(signs) <= 1``.
        return signs.pop()
    else:
        raise ValueError(
            'Did not reach a conclusion after max subdivisions',
            _MAX_POLY_SUBDIVISIONS)


def _2x2_det(mat):
    r"""Compute the determinant of a 2x2 matrix.

    This is "needed" because :func:`numpy.linalg.det` rounds off
    answers when it doesn't need to. For example:

    .. doctest:: 2-by-2

       >>> import numpy as np
       >>> mat = np.array([
       ...     [-24.0, 3.0],
       ...     [-27.0, 0.0],
       ... ]) / 16.0
       >>> actual_det = -mat[0, 1] * mat[1, 0]
       >>> np_det = np.linalg.det(mat)
       >>> actual_det == np_det
       False
       >>> abs(actual_det - np_det) < 1e-16
       True

    Args:
        mat (numpy.ndarray): A 2x2 matrix.

    Returns:
        float: The determinant of ``mat``.
    """
    return mat[0, 0] * mat[1, 1] - mat[0, 1] * mat[1, 0]


def quadratic_jacobian_polynomial(nodes):
    r"""Compute the Jacobian determinant of a quadratic surface.

    Converts :math:`\det(J(s, t))` to a polynomial on the reference
    triangle and represents it as a surface object.

    .. note::

       This assumes that ``nodes`` is 6x2 but doesn't verify this.
       (However, the multiplication by ``_QUADRATIC_JACOBIAN_HELPER``
       would fail if ``nodes`` wasn't 6xN and then the ensuing
       determinants would fail if there weren't 2 columns.)

    Args:
        nodes (numpy.ndarray): A 6x2 array of nodes in a surface.

    Returns:
        numpy.ndarray: Coefficients in Bernstein basis.
    """
    # First evaluate the Jacobian at each of the 6 nodes.
    # pylint: disable=no-member
    jac_parts = _QUADRATIC_JACOBIAN_HELPER.dot(nodes)
    # pylint: enable=no-member
    jac_at_nodes = np.empty((6, 1))
    jac_at_nodes[0, 0] = _2x2_det(jac_parts[:2, :])
    jac_at_nodes[1, 0] = _2x2_det(jac_parts[2:4, :])
    jac_at_nodes[2, 0] = _2x2_det(jac_parts[4:6, :])
    jac_at_nodes[3, 0] = _2x2_det(jac_parts[6:8, :])
    jac_at_nodes[4, 0] = _2x2_det(jac_parts[8:10, :])
    jac_at_nodes[5, 0] = _2x2_det(jac_parts[10:, :])

    # Convert the nodal values to the Bernstein basis...
    # pylint: disable=no-member
    bernstein = _QUADRATIC_TO_BERNSTEIN.dot(jac_at_nodes)
    # pylint: enable=no-member
    return bernstein


def cubic_jacobian_polynomial(nodes):
    r"""Compute the Jacobian determinant of a cubic surface.

    Converts :math:`\det(J(s, t))` to a polynomial on the reference
    triangle and represents it as a surface object.

    .. note::

       This assumes that ``nodes`` is 10x2 but doesn't verify this.
       (However, the multiplication by ``_CUBIC_JACOBIAN_HELPER``
       would fail if ``nodes`` wasn't 10xN and then the ensuing
       determinants would fail if there weren't 2 columns.)

    Args:
        nodes (numpy.ndarray): A 10x2 array of nodes in a surface.

    Returns:
        numpy.ndarray: Coefficients in Bernstein basis.
    """
    # First evaluate the Jacobian at each of the 15 nodes
    # in the quartic triangle.
    # pylint: disable=no-member
    jac_parts = _CUBIC_JACOBIAN_HELPER.dot(nodes)
    # pylint: enable=no-member
    jac_at_nodes = np.empty((15, 1))
    jac_at_nodes[0, 0] = _2x2_det(jac_parts[:2, :])
    jac_at_nodes[1, 0] = _2x2_det(jac_parts[2:4, :])
    jac_at_nodes[2, 0] = _2x2_det(jac_parts[4:6, :])
    jac_at_nodes[3, 0] = _2x2_det(jac_parts[6:8, :])
    jac_at_nodes[4, 0] = _2x2_det(jac_parts[8:10, :])
    jac_at_nodes[5, 0] = _2x2_det(jac_parts[10:12, :])
    jac_at_nodes[6, 0] = _2x2_det(jac_parts[12:14, :])
    jac_at_nodes[7, 0] = _2x2_det(jac_parts[14:16, :])
    jac_at_nodes[8, 0] = _2x2_det(jac_parts[16:18, :])
    jac_at_nodes[9, 0] = _2x2_det(jac_parts[18:20, :])
    jac_at_nodes[10, 0] = _2x2_det(jac_parts[20:22, :])
    jac_at_nodes[11, 0] = _2x2_det(jac_parts[22:24, :])
    jac_at_nodes[12, 0] = _2x2_det(jac_parts[24:26, :])
    jac_at_nodes[13, 0] = _2x2_det(jac_parts[26:28, :])
    jac_at_nodes[14, 0] = _2x2_det(jac_parts[28:, :])

    # Convert the nodal values to the Bernstein basis...
    # pylint: disable=no-member
    bernstein = _QUARTIC_TO_BERNSTEIN.dot(jac_at_nodes)
    # pylint: enable=no-member
    bernstein /= _QUARTIC_BERNSTEIN_FACTOR
    return bernstein


def de_casteljau_one_round(nodes, degree, lambda1, lambda2, lambda3):
    r"""Performs one "round" of the de Casteljau algorithm for surfaces.

    Converts the ``nodes`` into a basis for a surface one degree smaller
    by using the barycentric weights:

    .. math::

       q_{i, j, k} = \lambda_1 \cdot p_{i + 1, j, k} +
           \lambda_2 \cdot p_{i, j + 1, k} + \lambda_2 \cdot p_{i, j, k + 1}

    .. note:

       For degree :math:`d`d, the number of nodes should be
       :math:`(d + 1)(d + 2)/2`, but we don't verify this property.

    Args:
        nodes (numpy.ndarray): The nodes to reduce.
        degree (int): The degree of the surface.
        lambda1 (float): Parameter along the reference triangle.
        lambda2 (float): Parameter along the reference triangle.
        lambda3 (float): Parameter along the reference triangle.

    Returns:
        numpy.ndarray: The converted nodes.
    """
    num_nodes, dimension = nodes.shape
    num_new_nodes = num_nodes - degree - 1

    new_nodes = np.empty((num_new_nodes, dimension))

    index = 0
    # parent_i1 = index + k
    # parent_i2 = index + k + 1
    # parent_i3 = index + degree + 1
    parent_i1 = 0
    parent_i2 = 1
    parent_i3 = degree + 1
    for k in six.moves.xrange(degree):
        for unused_j in six.moves.xrange(degree - k):
            # NOTE: i = (degree - 1) - j - k
            new_nodes[index, :] = (
                lambda1 * nodes[parent_i1, :] +
                lambda2 * nodes[parent_i2, :] +
                lambda3 * nodes[parent_i3, :])
            # Update all the indices.
            parent_i1 += 1
            parent_i2 += 1
            parent_i3 += 1
            index += 1

        # Update the indices that depend on k.
        parent_i1 += 1
        parent_i2 += 1

    return new_nodes


def _make_transform(degree, weights_a, weights_b, weights_c):
    """Compute matrices corresponding to the de Casteljau algorithm.

    Applies the de Casteljau to the identity matrix, thus
    effectively caching the algorithm in a transformation matrix.

    .. note::

       This is premature optimization. It's unclear if the time
       saved from "caching" one round of de Casteljau is cancelled
       out by the extra storage required for the 3 matrices.

    Args:
        degree (int): The degree of a candidate surface.
        weights_a (Tuple[float, float, float]): Barycentric weights for a
            point in the reference triangle
        weights_b (Tuple[float, float, float]): Barycentric weights for a
            point in the reference triangle
        weights_c (Tuple[float, float, float]): Barycentric weights for a
            point in the reference triangle

    Returns:
        Mapping[int, numpy.ndarray]: Mapping from keys to the de Casteljau
        transformation mappings. The keys are ``0`` corresponding to
        ``weights_a``, ``1`` to ``weights_b`` and ``2`` to ``weights_c``.
    """
    num_nodes = ((degree + 1) * (degree + 2)) // 2
    id_mat = np.eye(num_nodes)

    # Pre-compute the matrices that do the reduction so we don't
    # have to **actually** perform the de Casteljau algorithm
    # every time.
    transform = {
        0: de_casteljau_one_round(id_mat, degree, *weights_a),
        1: de_casteljau_one_round(id_mat, degree, *weights_b),
        2: de_casteljau_one_round(id_mat, degree, *weights_c),
    }
    return transform


def _reduced_to_matrix(shape, degree, vals_by_weight):
    r"""Converts a reduced values dictionary into a matrix.

    The ``vals_by_weight`` mapping has keys of the form:
    ``(0, ..., 1, ..., 2, ...)`` where the ``0`` corresponds
    to the number of times the first set of barycentric
    weights was used in the reduction process, and similarly
    for ``1`` and ``2``.

    These points correspond to barycentric weights in their
    own right. For example ``(0, 0, 1, 2)`` corresponds to
    the Barycentric weight
    :math:`\left(\frac{2}{4}, \frac{1}{4}, \frac{1}{4}\right)`.

    Once the keys in ``vals_by_weight`` have been converted
    to Barycentric coordinates, we order them according to
    our rule (bottom to top, left to right) and then return
    them in a single matrix.

    Args:
        shape (tuple): The shape of the result matrix.
        degree (int): The degree of the surface.
        vals_by_weight (Mapping[tuple, numpy.ndarray]): Dictionary
            of reduced nodes according to blending of each of the
            three sets of weights in a reduction.

    Returns:
        numpy.ndarray: The newly created reduced control points.
    """
    result = np.empty(shape)
    index = 0
    for k in six.moves.xrange(degree + 1):
        for j in six.moves.xrange(degree + 1 - k):
            i = degree - j - k
            key = (0,) * i + (1,) * j + (2,) * k
            result[index, :] = vals_by_weight[key]
            index += 1

    return result


def specialize_surface(nodes, degree, weights_a, weights_b, weights_c):
    """Specialize a surface to a reparameterization

    Does so by taking three points (in Barycentric form) within the
    reference triangle and then reparameterizing the surface onto
    the triangle formed by those three points.

    .. note::

       This assumes the surface is degree 1 or greater but doesn't check.

    Args:
        nodes (numpy.ndarray): Control points for a surface.
        degree (int): The degree of the surface.
        weights_a (Tuple[float, float, float]): Barycentric weights for a
            point in the reference triangle
        weights_b (Tuple[float, float, float]): Barycentric weights for a
            point in the reference triangle
        weights_c (Tuple[float, float, float]): Barycentric weights for a
            point in the reference triangle

    Returns:
        numpy.ndarray: The control points for the specialized surface.
    """
    # Uses A-->0, B-->1, C-->2 to represent the specialization used.
    partial_vals = {
        (0,): de_casteljau_one_round(nodes, degree, *weights_a),
        (1,): de_casteljau_one_round(nodes, degree, *weights_b),
        (2,): de_casteljau_one_round(nodes, degree, *weights_c),
    }

    for reduced_deg in six.moves.xrange(degree - 1, 0, -1):
        new_partial = {}
        transform = _make_transform(
            reduced_deg, weights_a, weights_b, weights_c)
        for key, sub_nodes in six.iteritems(partial_vals):
            # Our keys are ascending so we increment from the last value.
            for next_id in six.moves.xrange(key[-1], 2 + 1):
                new_key = key + (next_id,)
                new_partial[new_key] = transform[next_id].dot(sub_nodes)

        partial_vals = new_partial

    return _reduced_to_matrix(nodes.shape, degree, partial_vals)


def _mean_centroid(candidates):
    """Take the mean of all centroids in set of reference triangles.

    Args:
        candidates (List[.Surface]): List of surfaces. We'll only use
            the base ``x`` and ``y`` and the width of the reference
            triangle that each surface represents.

    Returns:
        Tuple[float, float]: The mean of all centroids.
    """
    sum_x = 0.0
    sum_y = 0.0
    sum_width = 0.0
    for candidate in candidates:
        sum_x += candidate.base_x
        sum_y += candidate.base_y
        sum_width += candidate.width

    denom = float(len(candidates))
    mean_x = sum_x / denom
    mean_y = sum_y / denom
    width_delta = sum_width / (3.0 * denom)

    return mean_x + width_delta, mean_y + width_delta


def jacobian_s(nodes, degree, dimension):
    r"""Compute :math:`\frac{\partial B}{\partial s}`.

    Args:
        nodes (numpy.ndarray): Array of nodes in a surface.
        degree (int): The degree of the surface.
        dimension (int): The dimension the surface lives in.

    Returns:
        numpy.ndarray: Nodes of the Jacobian surface in
            B |eacute| zier form.
    """
    num_nodes = (degree * (degree + 1)) // 2
    result = np.empty((num_nodes, dimension))

    index = 0
    i = 0
    for num_vals in six.moves.xrange(degree, 0, -1):
        for _ in six.moves.xrange(num_vals):
            result[index, :] = nodes[i + 1, :] - nodes[i, :]
            # Update the indices
            index += 1
            i += 1

        # In between each row, the index gains an extra value.
        i += 1

    return float(degree) * result


def jacobian_t(nodes, degree, dimension):
    r"""Compute :math:`\frac{\partial B}{\partial t}`.

    Args:
        nodes (numpy.ndarray): Array of nodes in a surface.
        degree (int): The degree of the surface.
        dimension (int): The dimension the surface lives in.

    Returns:
        numpy.ndarray: Nodes of the Jacobian surface in
            B |eacute| zier form.
    """
    num_nodes = (degree * (degree + 1)) // 2
    result = np.empty((num_nodes, dimension))

    index = 0
    i = 0
    j = degree + 1
    for num_vals in six.moves.xrange(degree, 0, -1):
        for _ in six.moves.xrange(num_vals):
            result[index, :] = nodes[j, :] - nodes[i, :]
            # Update the indices
            index += 1
            i += 1
            j += 1

        # In between each row, the index gains an extra value.
        i += 1

    return float(degree) * result


def newton_refine(surface, x_val, y_val, s, t):
    r"""Refine a solution to :math:`B(s, t) = p` using Newton's method.

    Computes updates via

    .. math::

       \left[\begin{array}{c}
           0 \\ 0 \end{array}\right] \approx
           \left(B\left(s_{\ast}, t_{\ast}\right) -
           \left[\begin{array}{c} x \\ y \end{array}\right]\right) +
           \left[\begin{array}{c c}
               B_s\left(s_{\ast}, t_{\ast}\right) &
               B_t\left(s_{\ast}, t_{\ast}\right) \end{array}\right]
           \left[\begin{array}{c}
               \Delta s \\ \Delta t \end{array}\right]

    For example, (with weights
    :math:`\lambda_1 = 1 - s - t, \lambda_2 = s, \lambda_3 = t`)
    consider the surface

    .. math::

       B(s, t) =
           \left[\begin{array}{c} 0 \\ 0 \end{array}\right] \lambda_1^2 +
           \left[\begin{array}{c} 1 \\ 0 \end{array}\right]
               2 \lambda_1 \lambda_2 +
           \left[\begin{array}{c} 2 \\ 0 \end{array}\right] \lambda_2^2 +
           \left[\begin{array}{c} 2 \\ 1 \end{array}\right]
               2 \lambda_1 \lambda_3 +
           \left[\begin{array}{c} 2 \\ 2 \end{array}\right]
               2 \lambda_2 \lambda_1 +
           \left[\begin{array}{c} 0 \\ 2 \end{array}\right] \lambda_3^2

    and the point
    :math:`B\left(\frac{1}{4}, \frac{1}{2}\right) =
    \frac{1}{4} \left[\begin{array}{c} 5 \\ 5 \end{array}\right]`.

    Starting from the **wrong** point
    :math:`s = \frac{1}{2}, t = \frac{1}{4}`, we have

    .. math::

       \begin{align*}
       \left[\begin{array}{c} x \\ y \end{array}\right] -
          B\left(\frac{1}{2}, \frac{1}{4}\right) &= \frac{1}{4}
          \left[\begin{array}{c} -1 \\ 2 \end{array}\right] \\
       DB\left(\frac{1}{2}, \frac{1}{4}\right) &= \frac{1}{2}
           \left[\begin{array}{c c} 3 & 2 \\ 1 & 6 \end{array}\right] \\
       \Longrightarrow \left[\begin{array}{c}
           \Delta s \\ \Delta t \end{array}\right] &= \frac{1}{32}
           \left[\begin{array}{c} -10 \\ 7 \end{array}\right]
       \end{align*}

    .. image:: images/newton_refine_surface.png
       :align: center

    .. testsetup:: newton-refine-surface

       import numpy as np
       import bezier
       from bezier._surface_helpers import newton_refine

    .. doctest:: newton-refine-surface

       >>> surface = bezier.Surface(np.array([
       ...     [0.0, 0.0],
       ...     [1.0, 0.0],
       ...     [2.0, 0.0],
       ...     [2.0, 1.0],
       ...     [2.0, 2.0],
       ...     [0.0, 2.0],
       ... ]))
       >>> surface.is_valid
       True
       >>> x_val, y_val = surface.evaluate_cartesian(0.25, 0.5)
       >>> x_val, y_val
       (1.25, 1.25)
       >>> s, t = 0.5, 0.25
       >>> new_s, new_t = newton_refine(surface, x_val, y_val, s, t)
       >>> 32 * (new_s - s)
       -10.0
       >>> 32 * (new_t - t)
       7.0

    .. testcleanup:: newton-refine-surface

       import make_images
       make_images.newton_refine_surface(
           surface, x_val, y_val, s, t, new_s, new_t)

    Args:
        surface (.Surface): A B |eacute| zier surface (assumed to
            be two-dimensional).
        x_val (float): The :math:`x`-coordinate of a point
            on the surface.
        y_val (float): The :math:`y`-coordinate of a point
            on the surface.
        s (float): Approximate :math:`s`-value to be refined.
        t (float): Approximate :math:`t`-value to be refined.

    Returns:
        Tuple[float, float]: The refined :math:`s` and :math:`t` values.
    """
    surf_x, surf_y = surface.evaluate_cartesian(s, t)
    if surf_x == x_val and surf_y == y_val:
        # No refinement is needed.
        return s, t

    nodes = surface._nodes  # pylint: disable=protected-access
    # Compute Jacobian nodes / stack them horizatonally.
    jac_both = np.hstack([
        jacobian_s(nodes, surface.degree, surface.dimension),
        jacobian_t(nodes, surface.degree, surface.dimension),
    ])

    lambda1 = 1.0 - s - t
    # The degree of the jacobian is one less.
    for reduced_deg in six.moves.xrange(surface.degree - 1, 0, -1):
        jac_both = de_casteljau_one_round(
            jac_both, reduced_deg, lambda1, s, t)

    # The first column of the jacobian matrix is B_s (i.e. the
    # left-most values in ``jac_both``).
    jac_mat = np.array([
        [jac_both[0, 0], jac_both[0, 2]],
        [jac_both[0, 1], jac_both[0, 3]],
    ])
    rhs = np.array([
        [x_val - surf_x],
        [y_val - surf_y],
    ])
    soln = np.linalg.solve(jac_mat, rhs)
    return s + soln[0, 0], t + soln[1, 0]


def locate_point(surface, x_val, y_val):
    r"""Locate a point on a surface.

    Does so by recursively subdividing the surface and rejecting
    sub-surfaces with bounding boxes that don't contain the point.
    After the sub-surfaces are sufficiently small, uses Newton's
    method to zoom in on the intersection.

    Args:
        surface (.Surface): A B |eacute| zier surface (assumed to
            be two-dimensional).
        x_val (float): The :math:`x`-coordinate of a point
            on the surface.
        y_val (float): The :math:`y`-coordinate of a point
            on the surface.

    Returns:
        Optional[Tuple[float, float]]: The :math:`s` and :math:`t`
        values corresponding to ``x_val`` and ``y_val`` or
        :data:`None` if the point is not on the ``surface``.
    """
    candidates = [surface]
    for _ in six.moves.xrange(_MAX_LOCATE_SUBDIVISIONS + 1):
        next_candidates = []
        for candidate in candidates:
            nodes = candidate._nodes  # pylint: disable=protected-access
            if _helpers.contains(nodes, x_val, y_val):
                next_candidates.extend(candidate.subdivide())

        candidates = next_candidates

    if not candidates:
        return None

    # We take the average of all centroids from the candidates
    # that may contain the point.
    s_approx, t_approx = _mean_centroid(candidates)
    s, t = newton_refine(surface, x_val, y_val, s_approx, t_approx)

    actual = surface.evaluate_cartesian(s, t)
    expected = np.array([x_val, y_val])
    if not _helpers.vector_close(actual, expected, eps=_LOCATE_EPS):
        s, t = newton_refine(surface, x_val, y_val, s, t)
    return s, t


def classify_intersection(intersection):
    r"""Determine which curve is on the "inside of the intersection".

    This is intended to be a helper for forming a :class:`.CurvedPolygon`
    from the edge intersections of two :class:`.Surface`-s. In order
    to move from one intersection to another (or to the end of an edge),
    the interior edge must be determined at the point of intersection.

    The "typical" case is on the interior of both edges:

    .. image:: images/classify_intersection1.png
       :align: center

    .. testsetup:: classify-intersection1, classify-intersection2,
                   classify-intersection3, classify-intersection4,
                   classify-intersection5, classify-intersection6,
                   classify-intersection7

       import numpy as np
       import bezier
       from bezier import _curve_helpers
       from bezier._intersection_helpers import Intersection
       from bezier._surface_helpers import classify_intersection

       def hodograph(curve, s):
           return _curve_helpers.evaluate_hodograph(
               curve._nodes, curve.degree, s)

       def curvature(curve, s):
           nodes = curve._nodes
           tangent = _curve_helpers.evaluate_hodograph(
               nodes, curve.degree, s)
           return _curve_helpers.get_curvature(
               nodes, curve.degree, tangent, s)

    .. doctest:: classify-intersection1
       :options: +NORMALIZE_WHITESPACE

       >>> curve1 = bezier.Curve(np.array([
       ...     [1.0 , 0.0 ],
       ...     [1.75, 0.25],
       ...     [2.0 , 1.0 ],
       ... ]))
       >>> curve2 = bezier.Curve(np.array([
       ...     [0.0   , 0.0   ],
       ...     [1.6875, 0.0625],
       ...     [2.0   , 0.5   ],
       ... ]))
       >>> s, t = 0.25, 0.5
       >>> curve1.evaluate(s) == curve2.evaluate(t)
       array([ True, True], dtype=bool)
       >>> tangent1 = hodograph(curve1, s)
       >>> tangent1
       array([ 1.25, 0.75])
       >>> tangent2 = hodograph(curve2, t)
       >>> tangent2
       array([ 2. , 0.5])
       >>> intersection = Intersection(curve1, s, curve2, t)
       >>> classify_intersection(intersection)
       <IntersectionClassification.first: 'first'>

    .. testcleanup:: classify-intersection1

       import make_images
       make_images.classify_intersection1(
           s, curve1, tangent1, curve2, tangent2)

    We determine the interior / left one by using the `right-hand rule`_:
    by embedding the tangent vectors in :math:`\mathbf{R}^3`, we
    compute

    .. _right-hand rule: https://en.wikipedia.org/wiki/Right-hand_rule

    .. math::

       \left[\begin{array}{c}
           x_1'(s) \\ y_1'(s) \\ 0 \end{array}\right] \times
       \left[\begin{array}{c}
           x_2'(t) \\ y_2'(t) \\ 0 \end{array}\right] =
       \left[\begin{array}{c}
           0 \\ 0 \\ x_1'(s) y_2'(t) - x_2'(t) y_1'(s) \end{array}\right].

    If the cross-product quantity
    :math:`B_1'(s) \times B_2'(t) = x_1'(s) y_2'(t) - x_2'(t) y_1'(s)`
    is positive, then the first curve is "to the right", i.e. the second
    curve is interior. If the cross-product is negative, the first
    curve is interior.

    When :math:`B_1'(s) \times B_2'(t) = 0`, the tangent
    vectors are parallel, i.e. the intersection is a point of tangency:

    .. image:: images/classify_intersection2.png
       :align: center

    .. doctest:: classify-intersection2
       :options: +NORMALIZE_WHITESPACE

       >>> curve1 = bezier.Curve(np.array([
       ...     [1.0, 0.0],
       ...     [1.5, 1.0],
       ...     [2.0, 0.0],
       ... ]))
       >>> curve2 = bezier.Curve(np.array([
       ...     [0.0, 0.0],
       ...     [1.5, 1.0],
       ...     [3.0, 0.0],
       ... ]))
       >>> s, t = 0.5, 0.5
       >>> curve1.evaluate(s) == curve2.evaluate(t)
       array([ True, True], dtype=bool)
       >>> intersection = Intersection(curve1, s, curve2, t)
       >>> classify_intersection(intersection)
       <IntersectionClassification.tangent_second: 'tangent_second'>

    .. testcleanup:: classify-intersection2

       import make_images
       make_images.classify_intersection2(s, curve1, curve2)

    Depending on the direction of the parameterizations, the interior
    curve may change, but we can use the (signed) `curvature`_ of each
    curve at that point to determine which is on the interior:

    .. _curvature: https://en.wikipedia.org/wiki/Curvature

    .. image:: images/classify_intersection3.png
       :align: center

    .. doctest:: classify-intersection3
       :options: +NORMALIZE_WHITESPACE

       >>> curve1 = bezier.Curve(np.array([
       ...     [2.0, 0.0],
       ...     [1.5, 1.0],
       ...     [1.0, 0.0],
       ... ]))
       >>> curve2 = bezier.Curve(np.array([
       ...     [3.0, 0.0],
       ...     [1.5, 1.0],
       ...     [0.0, 0.0],
       ... ]))
       >>> s, t = 0.5, 0.5
       >>> curve1.evaluate(s) == curve2.evaluate(t)
       array([ True, True], dtype=bool)
       >>> intersection = Intersection(curve1, s, curve2, t)
       >>> classify_intersection(intersection)
       <IntersectionClassification.tangent_first: 'tangent_first'>

    .. testcleanup:: classify-intersection3

       import make_images
       make_images.classify_intersection3(s, curve1, curve2)

    When the curves are moving in opposite directions at a point
    of tangency, there is no side to choose. Either the point of tangency
    is not part of any :class:`.CurvedPolygon` intersection

    .. image:: images/classify_intersection4.png
       :align: center

    .. doctest:: classify-intersection4
       :options: +NORMALIZE_WHITESPACE

       >>> curve1 = bezier.Curve(np.array([
       ...     [2.0, 0.0],
       ...     [1.5, 1.0],
       ...     [1.0, 0.0],
       ... ]))
       >>> curve2 = bezier.Curve(np.array([
       ...     [0.0, 0.0],
       ...     [1.5, 1.0],
       ...     [3.0, 0.0],
       ... ]))
       >>> s, t = 0.5, 0.5
       >>> curve1.evaluate(s) == curve2.evaluate(t)
       array([ True, True], dtype=bool)
       >>> intersection = Intersection(curve1, s, curve2, t)
       >>> classify_intersection(intersection)
       <IntersectionClassification.opposed: 'opposed'>

    .. testcleanup:: classify-intersection4

       import make_images
       make_images.classify_intersection4(s, curve1, curve2)

    or the point of tangency is a "degenerate" part of two
    :class:`.CurvedPolygon` intersections. It is "degenerate"
    because from one direction, the point should be classified
    as ``1`` and from another as ``0``:

    .. image:: images/classify_intersection5.png
       :align: center

    .. doctest:: classify-intersection5
       :options: +NORMALIZE_WHITESPACE

       >>> curve1 = bezier.Curve(np.array([
       ...     [1.0, 0.0],
       ...     [1.5, 1.0],
       ...     [2.0, 0.0],
       ... ]))
       >>> curve2 = bezier.Curve(np.array([
       ...     [3.0, 0.0],
       ...     [1.5, 1.0],
       ...     [0.0, 0.0],
       ... ]))
       >>> s, t = 0.5, 0.5
       >>> curve1.evaluate(s) == curve2.evaluate(t)
       array([ True, True], dtype=bool)
       >>> intersection = Intersection(curve1, s, curve2, t)
       >>> classify_intersection(intersection)
       Traceback (most recent call last):
         ...
       NotImplementedError: Curves moving in opposite direction
                            but define overlapping arcs.

    .. testcleanup:: classify-intersection5

       import make_images
       make_images.classify_intersection5(s, curve1, curve2)

    However, if the `curvature`_ of each curve is identical, we
    don't try to distinguish further:

    .. image:: images/classify_intersection6.png
       :align: center

    .. doctest:: classify-intersection6
       :options: +NORMALIZE_WHITESPACE

       >>> curve1 = bezier.Curve(np.array([
       ...     [ 0.375,  0.0625],
       ...     [-0.125, -0.0625],
       ...     [-0.125,  0.0625],
       ... ]))
       >>> curve2 = bezier.Curve(np.array([
       ...     [ 0.75,  0.25],
       ...     [-0.25, -0.25],
       ...     [-0.25,  0.25],
       ... ]))
       >>> s, t = 0.5, 0.5
       >>> curve1.evaluate(s) == curve2.evaluate(t)
       array([ True, True], dtype=bool)
       >>> hodograph(curve1, s)
       array([-0.5, 0. ])
       >>> hodograph(curve2, t)
       array([-1., 0.])
       >>> curvature(curve1, s)
       -2.0
       >>> curvature(curve2, t)
       -2.0
       >>> intersection = Intersection(curve1, s, curve2, t)
       >>> classify_intersection(intersection)
       Traceback (most recent call last):
         ...
       NotImplementedError: Tangent curves have same curvature.

    .. testcleanup:: classify-intersection6

       import make_images
       make_images.classify_intersection6(s, curve1, curve2)

    In addition to points of tangency, intersections that happen at
    the end of an edge need special handling:

    .. image:: images/classify_intersection7.png
       :align: center

    .. doctest:: classify-intersection7
       :options: +NORMALIZE_WHITESPACE

       >>> curve1a = bezier.Curve(np.array([
       ...     [0.0, 0.0 ],
       ...     [4.5, 0.0 ],
       ...     [9.0, 2.25],
       ... ]))
       >>> curve2 = bezier.Curve(np.array([
       ...     [11.25, 0.0],
       ...     [ 9.0 , 4.5],
       ...     [ 2.75, 1.0],
       ... ]))
       >>> s, t = 1.0, 0.375
       >>> curve1a.evaluate(s) == curve2.evaluate(t)
       array([ True, True], dtype=bool)
       >>> intersection = Intersection(curve1a, s, curve2, t)
       >>> classify_intersection(intersection)
       Traceback (most recent call last):
         ...
       ValueError: ('Intersection occurs at the end of an edge',
                    's', 1.0, 't', 0.375)
       >>>
       >>> curve1b = bezier.Curve(np.array([
       ...     [9.0, 2.25 ],
       ...     [4.5, 2.375],
       ...     [0.0, 2.5  ],
       ... ]))
       >>> curve1b.evaluate(0.0) == curve2.evaluate(t)
       array([ True, True], dtype=bool)
       >>> intersection = Intersection(curve1b, 0.0, curve2, t)
       >>> classify_intersection(intersection)
       <IntersectionClassification.first: 'first'>

    .. testcleanup:: classify-intersection7

       import make_images
       make_images.classify_intersection7(s, curve1a, curve1b, curve2)

    .. note::

       This assumes the intersection occurs in :math:`\mathbf{R}^2`
       but doesn't check this.

    .. note::

       This function doesn't allow wiggle room / round-off when checking
       endpoints, nor when checking if the cross-product is near zero,
       nor when curvatures are compared. However, the most "correct"
       version of this function likely should allow for some round off.

    Args:
        intersection (.Intersection): An intersection object.

    Returns:
        IntersectionClassification: The "inside" curve type, based on
        the classification enum.

    Raises:
        ValueError: If the intersection occurs at the end of either
            curve involved. This is because we want to classify which
            curve to **move forward** on, and we can't move past the
            end of a segment.
    """
    if intersection.s == 1.0 or intersection.t == 1.0:
        raise ValueError('Intersection occurs at the end of an edge',
                         's', intersection.s, 't', intersection.t)

    tangent1 = _curve_helpers.evaluate_hodograph(
        intersection.left._nodes,  # pylint: disable=protected-access
        intersection.left.degree, intersection.s)
    tangent2 = _curve_helpers.evaluate_hodograph(
        intersection.right._nodes,  # pylint: disable=protected-access
        intersection.right.degree, intersection.t)

    # Take the cross-product of tangent vectors to determine which one
    # is more "to the left".
    cross_prod = _helpers.cross_product(
        np.array([tangent1]), np.array([tangent2]))
    if cross_prod < 0:
        return IntersectionClassification.first
    elif cross_prod > 0:
        return IntersectionClassification.second
    else:
        return _classify_tangent_intersection(
            intersection, tangent1, tangent2)


def _classify_tangent_intersection(intersection, tangent1, tangent2):
    """Helper for func:`classify_intersection` at tangencies.

    Args:
        intersection (.Intersection): An intersection object.
        tangent1 (numpy.ndarray): The tangent vector on the ``left`` curve
            at the intersection.
        tangent2 (numpy.ndarray): The tangent vector on the ``right`` curve
            at the intersection.

    Returns:
        IntersectionClassification: The "inside" curve type, based on
        the classification enum. Will either be ``opposed`` or one
        of the ``tangent`` values.

    Raises:
        NotImplementedError: If the curves are tangent, moving in opposite
            directions, but enclose overlapping arcs.
        NotImplementedError: If the curves are tangent at the intersection
            and have the same curvature.
    """
    dot_prod = tangent1.dot(tangent2)
    # NOTE: When computing curvatures we assume that we don't have lines
    #       here, because lines that are tangent at an intersection are
    #       parallel and we don't handle that case.
    curvature1 = _curve_helpers.get_curvature(
        intersection.left._nodes,  # pylint: disable=protected-access
        intersection.left.degree, tangent1, intersection.s)
    curvature2 = _curve_helpers.get_curvature(
        intersection.right._nodes,  # pylint: disable=protected-access
        intersection.right.degree, tangent2, intersection.t)
    if dot_prod < 0:
        # If the tangent vectors are pointing in the opposite direction,
        # then the curves are facing opposite directions.
        sign1, sign2 = _SIGN([curvature1, curvature2])
        if sign1 == sign2:
            # If both curvatures are positive, since the curves are
            # moving in opposite directions, the tangency isn't part of
            # the surface intersection.
            if sign1 == 1.0:
                return IntersectionClassification.opposed
            else:
                raise NotImplementedError(_BAD_TANGENT)
        else:
            delta_c = abs(curvature1) - abs(curvature2)
            if delta_c == 0.0:
                raise NotImplementedError(_SAME_CURVATURE)
            elif sign1 == _SIGN(delta_c):
                return IntersectionClassification.opposed
            else:
                raise NotImplementedError(_BAD_TANGENT)
    else:
        if curvature1 > curvature2:
            return IntersectionClassification.tangent_first
        elif curvature1 < curvature2:
            return IntersectionClassification.tangent_second
        else:
            raise NotImplementedError(_SAME_CURVATURE)


def edge_cycle(edge1, edge2, edge3):
    """Make edges follow the cycle ``1->2->3->1``.

    Does this by setting the "previous" and "next" edge
    values on each curved edge.

    Args:
        edge1 (.Curve): First curve in cycle.
        edge2 (.Curve): Second curve in cycle.
        edge3 (.Curve): Third curve in cycle.
    """
    # pylint: disable=protected-access
    edge1._edge_index = 0
    edge1._next_edge = edge2
    edge1._previous_edge = edge3

    edge2._edge_index = 1
    edge2._next_edge = edge3
    edge2._previous_edge = edge1

    edge3._edge_index = 2
    edge3._next_edge = edge1
    edge3._previous_edge = edge2
    # pylint: enable=protected-access


def handle_corners(intersection):
    """Mutates an intersection if it is on a corner.

    Does nothing if the intersection happens in the middle of two
    edges.

    If the intersection occurs at the end of the ``left`` curve,
    moves it to the beginning of the next edge. Similar for the
    ``right`` curve.

    This function is used as a pre-processing step before passing
    an intersection to :func:`classify_intersection`. There, only
    corners that **begin** an edge are considered, since that
    function is trying to determine which edge to **move forward** on.

    .. note::

       This assumes the ``left`` and ``right`` curves are edges
       in a surface, so the code (may) rely on ``next_edge``
       and / or ``previous_edge`` being valid.

    Args:
        intersection (.Intersection): An intersection to mutate.

    Returns:
        bool: Indicating if the object was changed.
    """
    changed = False
    if intersection.s == 1.0:
        # pylint: disable=protected-access
        intersection._s_val = 0.0
        intersection._left = intersection.left.next_edge
        # pylint: enable=protected-access
        changed = True
    if intersection.t == 1.0:
        # pylint: disable=protected-access
        intersection._t_val = 0.0
        intersection._right = intersection.right.next_edge
        # pylint: enable=protected-access
        changed = True

    return changed


def _identifier(intersection):
    """Provide a simple value to identify an intersection.

    Args:
        intersection (.Intersection): The current intersection.

    Returns:
        Tuple[int, float, int, float]: The edge indices (left first,
        right third) for the intersection and the parameter values
        (left second, right fourth).
    """
    return (intersection.left.edge_index, intersection.s,
            intersection.right.edge_index, intersection.t)


def verify_duplicates(duplicates, uniques):
    """Verify that a set of intersections had expected duplicates.

    .. note::

       This method only considers two intersections as equal if the
       ``s`` and ``t`` values are bit-wise identical.

    Args:
        duplicates (List[.Intersection]): List of intersections
            corresponding to duplicates that were filtered out.
        uniques (List[.Intersection]): List of "final" intersections
            with duplicates filtered out.

    Raises:
        ValueError: If the ``uniques`` are not actually all unique.
        ValueError: If one of the ``duplicates`` does not correspond to
            an intersection in ``uniques``.
        ValueError: If a duplicate occurs only once but does not have
            exactly one of ``s`` and ``t`` equal to ``0.0``.
        ValueError: If a duplicate occurs three times but does not have
            exactly both ``s == t == 0.0``.
        ValueError: If a duplicate occurs a number other than one or three
            times.
    """
    counter = collections.Counter(_identifier(dupe) for dupe in duplicates)
    uniques_keys = set(_identifier(uniq) for uniq in uniques)
    if len(uniques_keys) < len(uniques):
        raise ValueError('Non-unique intersection')

    for key, count in six.iteritems(counter):
        if key not in uniques_keys:
            raise ValueError('Duplicate not among uniques', key)

        if count == 1:
            if (key[1], key[3]).count(0.0) != 1:
                raise ValueError('Count == 1 should be a single corner', key)
        elif count == 3:
            if (key[1], key[3]) != (0.0, 0.0):
                raise ValueError('Count == 3 should be a double corner', key)
        else:
            raise ValueError('Unexpected duplicate count', count)


def _to_front(intersection, intersections, unused):
    """Rotates a node to the "front".

    Helper for :func:`combine_intersections`.

    If a node is at the end of a segment, moves it to the beginning
    of the next segment (at the exact same point).

    .. note::

        This method checks for **exact** endpoints, i.e. parameter
        bitwise identical to ``1.0``. But we should probably allow
        some wiggle room.

    Args:
        intersection (.Intersection): The current intersection.
        intersections (List[.Intersection]): List of all detected
            intersections, provided as a reference for potential
            points to arrive at.
        unused (List[.Intersection]): List of nodes that haven't been
            used yet in an intersection curved polygon

    Returns:
        .Intersection: An intersection to (maybe) move to the beginning
        of the next segment(s).
    """
    changed = False
    if intersection.s == 1.0:
        changed = True
        new_intersection = _intersection_helpers.Intersection(
            intersection.left.next_edge, 0.0,
            intersection.right, intersection.t)
        new_intersection.interior_curve = intersection.interior_curve
        intersection = new_intersection

    if intersection.t == 1.0:
        changed = True
        new_intersection = _intersection_helpers.Intersection(
            intersection.left, intersection.s,
            intersection.right.next_edge, 0.0)
        new_intersection.interior_curve = intersection.interior_curve
        intersection = new_intersection

    if changed:
        # Make sure we haven't accidentally ignored an existing intersection.
        for other_int in intersections:
            if (other_int.s == intersection.s and
                    other_int.left is intersection.left):
                intersection = other_int
                break

            if (other_int.t == intersection.t and
                    other_int.right is intersection.right):
                intersection = other_int
                break

    if intersection in unused:
        unused.remove(intersection)
    return intersection


# pylint: disable=too-many-branches
def _get_next(intersection, intersections, unused):
    """Gets the next node along a given edge.

    Helper for :func:`combine_intersections`, but does the majority
    of the heavy lifting.

    .. note::

        This function returns :class:`.Intersection` objects even
        when the point isn't strictly an intersection. This is
        "incorrect" in some sense, but for now, we don't bother
        implementing a class similar to, but different from,
        :class:`.Intersection` to satisfy this need.

    Args:
        intersection (.Intersection): The current intersection.
        intersections (List[.Intersection]): List of all detected
            intersections, provided as a reference for potential
            points to arrive at.
        unused (List[.Intersection]): List of nodes that haven't been
            used yet in an intersection curved polygon

    Returns:
        .Intersection: The "next" point along a surface of intersection.
        This will produce the next intersection along the current edge or
        the end of the current edge.

    Raises:
        ValueError: If the intersection is not classified as
            :attr:`~.IntersectionClassification.first` or
            :attr:`~.IntersectionClassification.second`.
    """
    acceptable = (IntersectionClassification.first,
                  IntersectionClassification.second)

    result = None
    if intersection.interior_curve is IntersectionClassification.first:
        along_edge = None
        left = intersection.left
        s = intersection.s
        for other_int in intersections:
            other_s = other_int.s
            if other_int.left is left and other_s > s:
                # NOTE: We skip tangent intersections that don't occur
                #       at a corner.
                if (other_s < 1.0 and
                        other_int.interior_curve not in acceptable):
                    continue
                if along_edge is None or other_s < along_edge.s:
                    along_edge = other_int

        if along_edge is None:
            # Just return the segment end.
            new_intersection = _intersection_helpers.Intersection(
                intersection.left, 1.0, None, None)
            new_intersection.interior_curve = IntersectionClassification.first
            result = new_intersection
        else:
            result = along_edge
    elif intersection.interior_curve is IntersectionClassification.second:
        along_edge = None
        right = intersection.right
        t = intersection.t
        for other_int in intersections:
            other_t = other_int.t
            if other_int.right is right and other_t > t:
                # NOTE: We skip tangent intersections that don't occur
                #       at a corner.
                if (other_t < 1.0 and
                        other_int.interior_curve not in acceptable):
                    continue
                if along_edge is None or other_t < along_edge.t:
                    along_edge = other_int

        if along_edge is None:
            # Just return the segment end.
            new_intersection = _intersection_helpers.Intersection(
                None, None, intersection.right, 1.0)
            # pylint: disable=redefined-variable-type
            new_intersection.interior_curve = IntersectionClassification.second
            # pylint: enable=redefined-variable-type
            result = new_intersection
        else:
            result = along_edge
    else:
        raise ValueError('Cannot get next node if not starting from '
                         '"first" or "second".')

    if result in unused:
        unused.remove(result)
    return result
# pylint: enable=too-many-branches


def _ends_to_curve(start_node, end_node):
    """Convert a "pair" of intersection nodes to a curve segment.

    .. note::

       This function determines "left" or "right" curve based on the
       classification of ``start_node``, but the callers of this
       function could provide that information / isolate the
       base curve and the two parameters for us.

    .. note::

       This only checks the classification of the ``start_node``.

    Args:
        start_node (.Intersection): The beginning of a segment.
        end_node (.Intersection): The end of (the same) segment.

    Returns:
        .Curve: The segment between the nodes.

    Raises:
        ValueError: If the ``start_node`` and ``end_node`` disagree on
            the "left" curve when classified as "first" or disagree on
            the "right" curve when classified as "second".
        ValueError: If the ``start_node`` is not classified as
            :attr:`~.IntersectionClassification.first` or
            :attr:`~.IntersectionClassification.second`.
    """
    if start_node.interior_curve is IntersectionClassification.first:
        left = start_node.left
        if end_node.left is not left:
            raise ValueError(_WRONG_CURVE)
        return left.specialize(start_node.s, end_node.s)
    elif start_node.interior_curve is IntersectionClassification.second:
        right = start_node.right
        if end_node.right is not right:
            raise ValueError(_WRONG_CURVE)
        return right.specialize(start_node.t, end_node.t)
    else:
        raise ValueError('Segment start must be classified as '
                         '"first" or "second".')


def _to_curved_polygon(surface):
    """Convert a surface to a curved polygon.

    This is a helper for :func:`combine_intersections` to allow
    returning a curved polygon when the intersection is the
    entirety of one surface.

    Args:
        surface (.Surface): The surface to convert.

    Returns:
        .CurvedPolygon: The converted object.
    """
    edges = surface._get_edges()  # pylint: disable=protected-access
    return curved_polygon.CurvedPolygon(*edges)


def _no_intersections(surface1, surface2):
    """Determine if one surface is in the other.

    Helper for :func:`combine_intersections` that handles the case
    of no points of intersection. In this case, either the surfaces
    are disjoint or one is fully contained in the other.

    To check containment, it's enough to check if one of the corners
    is contained in the other surface.

    Args:
        surface1 (.Surface): First surface in intersection.
        surface2 (.Surface): Second surface in intersection.

    Returns:
        list: Either an empty list if one surface isn't contained
        in the other. Otherwise, the list will have a single
        :class:`.CurvedPolygon` corresponding to the internal surface.
    """
    nodes1 = surface1._nodes  # pylint: disable=protected-access
    corner1 = nodes1[[0], :]
    if surface2.locate(corner1) is not None:
        return [_to_curved_polygon(surface1)]

    nodes2 = surface2._nodes  # pylint: disable=protected-access
    corner2 = nodes2[[0], :]
    if surface1.locate(corner2) is not None:
        return [_to_curved_polygon(surface2)]

    return []


def _tangent_only_intersections(intersections, surface1, surface2):
    """Determine intersection in the case of only-tangent intersections.

    If the only intersections are tangencies, then either the surfaces
    are tangent but don't meet ("kissing" edges) or one surface is
    internally tangent to the other.

    Thus we expect every intersection in ``intersections`` to be
    classified as :attr:`~.IntersectionClassification.tangent_first`,
    :attr:`~.IntersectionClassification.tangent_second` or
    :attr:`~.IntersectionClassification.opposed`.

    What's more, we expect all intersections to be classified the same for
    a given pairing.

    Args:
        intersections (list): A list of :class:`.Intersection` objects
            produced by :func:`.all_intersections` applied to each of
            the 9 edge-edge pairs from a surface-surface pairing.
        surface1 (.Surface): First surface in intersection.
        surface2 (.Surface): Second surface in intersection.

    Returns:
        list: Either an empty list if one surface isn't contained
        in the other. Otherwise, the list will have a single
        :class:`.CurvedPolygon` corresponding to the internal surface.

    Raises:
        ValueError: If there are intersections of more than one type among
            :attr:`~.IntersectionClassification.tangent_first`,
            :attr:`~.IntersectionClassification.tangent_second` or
            :attr:`~.IntersectionClassification.opposed`.
        ValueError: If there is a unique classification, but it isn't one
            of the tangent types.
    """
    all_types = set([intersection.interior_curve
                     for intersection in intersections])
    if len(all_types) != 1:
        raise ValueError('Unexpected value, types should all match',
                         all_types)
    point_type = all_types.pop()
    if point_type is IntersectionClassification.opposed:
        return []
    elif point_type is IntersectionClassification.tangent_first:
        return [_to_curved_polygon(surface1)]
    elif point_type is IntersectionClassification.tangent_second:
        return [_to_curved_polygon(surface2)]
    else:
        raise ValueError('Point type not for tangency', point_type)


def _basic_interior_combine(intersections):
    """Combine intersections that don't involve tangencies.

    Helper for :func:`combine_intersections`.

    .. note::

       This helper assumes ``intersections`` isn't empty, but doesn't
       enforce it.

    Args:
        intersections (list): A list of :class:`.Intersection` objects
            produced by :func:`.all_intersections` applied to each of
            the 9 edge-edge pairs from a surface-surface pairing.

    Returns:
        List[.CurvedPolygon]: All of the intersections encountered.
    """
    acceptable = (IntersectionClassification.first,
                  IntersectionClassification.second)
    unused = [intersection for intersection in intersections
              if intersection.interior_curve in acceptable]
    result = []
    while unused:
        start = unused.pop()
        curr_node = start
        next_node = _get_next(start, intersections, unused)
        edge_ends = [(curr_node, next_node)]
        while next_node is not start:
            curr_node = _to_front(next_node, intersections, unused)
            # NOTE: We also check to break when moving a corner node
            #       to the front. This is because ``intersections``
            #       de-duplicates corners by selecting the one
            #       (of 2 or 4 choices) at the front of segment(s).
            if curr_node is start:
                break
            next_node = _get_next(curr_node, intersections, unused)
            edge_ends.append((curr_node, next_node))

        result.append(curved_polygon.CurvedPolygon(*(
            _ends_to_curve(*pair)
            for pair in edge_ends
        )))

    return result


def combine_intersections(intersections, surface1, surface2):
    """Combine curve-curve intersections into curved polygon(s).

    Does so assuming each intersection lies on an edge of one of
    two :class:`.Surface`-s.

    .. note ::

       This assumes that each ``intersection`` has been classified via
       :func:`classify_intersection`.

    Args:
        intersections (list): A list of :class:`.Intersection` objects
            produced by :func:`.all_intersections` applied to each of
            the 9 edge-edge pairs from a surface-surface pairing.
        surface1 (.Surface): First surface in intersection.
        surface2 (.Surface): Second surface in intersection.

    Returns:
        List[~bezier.curved_polygon.CurvedPolygon]: A list of curved polygons
        that compose the intersected objects.
    """
    if len(intersections) == 0:
        return _no_intersections(surface1, surface2)

    result = _basic_interior_combine(intersections)
    if len(result) > 0:
        return result

    return _tangent_only_intersections(intersections, surface1, surface2)


class IntersectionClassification(enum.Enum):
    """Enum classifying the "interior" curve in an intersection."""
    first = 'first'
    second = 'second'
    opposed = 'opposed'
    tangent_first = 'tangent_first'
    tangent_second = 'tangent_second'



# See the License for the specific language governing permissions and
# limitations under the License.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import mock


class Test_add_plot_boundary(unittest.TestCase):

    AX_PROPS = ('lines', 'set_xlim', 'set_ylim')

    @staticmethod
    def _call_function_under_test(ax, **kwargs):
        from bezier import _plot_helpers

        return _plot_helpers.add_plot_boundary(ax, **kwargs)

    def test_default(self):
        import matplotlib.lines

        line = matplotlib.lines.Line2D([-1.0, 1.0], [-1.0, 1.0])
        ax = mock.Mock(name='AxesSubplot', spec=self.AX_PROPS, lines=[line])

        self.assertIsNone(self._call_function_under_test(ax))
        ax.set_xlim.assert_called_once_with(-1.125, 1.125)
        ax.set_ylim.assert_called_once_with(-1.125, 1.125)

    def test_with_padding(self):
        import matplotlib.lines

        line = matplotlib.lines.Line2D([-1.0, 1.0], [-1.0, 1.0])
        ax = mock.Mock(name='AxesSubplot', spec=self.AX_PROPS, lines=[line])

        self.assertIsNone(self._call_function_under_test(ax, padding=0.5))
        ax.set_xlim.assert_called_once_with(-1.5, 1.5)
        ax.set_ylim.assert_called_once_with(-1.5, 1.5)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

import collections

import matplotlib.pyplot as plt
import numpy as np
import pytest
import six

import bezier
from bezier import _plot_helpers

import runtime_utils


PARALLEL_FAILURE = ('Line segments parallel.',)
BAD_TANGENT = (
    'Curves moving in opposite direction but define '
    'overlapping arcs.')
TANGENT_FAILURE = 'The number of candidate intersections is too high.'
CONFIG = runtime_utils.Config()

# F1L = sympy.Matrix([[s, t]])
SURFACE1L = bezier.Surface(np.array([
    [0.0, 0.0],
    [1.0, 0.0],
    [0.0, 1.0],
]))
# F2L = sympy.Matrix([[
#     (9 * s + 2 * t - 1) / 8,
#     (9 * s + 7 * t - 1) / 16,
# ]])
SURFACE2L = bezier.Surface(np.array([
    [-0.125, -0.0625],
    [1.0, 0.5],
    [0.125, 0.375],
]))
# F3L = sympy.Matrix([[(4 * s + 1) / 4, (8 * t - 1) / 8]])
SURFACE3L = bezier.Surface(np.array([
    [0.25, -0.125],
    [1.25, -0.125],
    [0.25, 0.875],
]))
# F4L = sympy.Matrix([[5 * (2 * s + t - 1) / 4, 35 * t / 16]])
SURFACE4L = bezier.Surface(np.array([
    [-1.25, 0.0],
    [1.25, 0.0],
    [0.0, 2.1875],
]))
# F5L = sympy.Matrix([[
#     (21 * s - 21 * t + 4) / 8,
#     (21 * s + 21 * t - 5) / 32,
# ]])
SURFACE5L = bezier.Surface(np.array([
    [0.5, -0.15625],
    [3.125, 0.5],
    [-2.125, 0.5],
]))
# F6L = sympy.Matrix([[(1 - 4 * s) / 4, (1 - 4 * t) / 4]])
SURFACE6L = bezier.Surface(np.array([
    [0.25, 0.25],
    [-0.75, 0.25],
    [0.25, -0.75],
]))

# F1Q = sympy.Matrix([[
#     (2 * s - t**2 + t) / 2,
#     (s**2 + 2 * s * t - s + 2 * t) / 2,
# ]])
SURFACE1Q = bezier.Surface(np.array([
    [0.0, 0.0],
    [0.5, -0.25],
    [1.0, 0.0],
    [0.25, 0.5],
    [0.75, 0.75],
    [0.0, 1.0],
]))
# F2Q = sympy.Matrix([[
#     (3 * s**2 + 6 * s * t + 5 * s + 8 * t - 2) / 8,
#     (3 * s**2 - 3 * t**2 - 11 * s + 3 * t + 6) / 8,
# ]])
SURFACE2Q = bezier.Surface(np.array([
    [-0.25, 0.75],
    [0.0625, 0.0625],
    [0.75, -0.25],
    [0.25, 0.9375],
    [0.9375, 0.25],
    [0.75, 0.75],
]))
# F3Q = sympy.Matrix([[
#     (11 * s * t + 5 * t**2 + 8 * s + 3 * t) / 8,
#     -(4 * s**2 + 7 * s * t + 7 * t**2 - 4 * s - 15 * t) / 8,
# ]])
SURFACE3Q = bezier.Surface(np.array([
    [0.0, 0.0],
    [0.5, 0.25],
    [1.0, 0.0],
    [0.1875, 0.9375],
    [1.375, 0.75],
    [1.0, 1.0],
]))
# F4Q = sympy.Matrix([[
#     -(2 * s * t + t**2 - 16 * s - 4 * t) / 16,
#     (2 * s**2 + 2 * s * t - 2 * s + 3 * t + 1) / 4,
# ]])
SURFACE4Q = bezier.Surface(np.array([
    [0.0, 0.25],
    [0.5, 0.0],
    [1.0, 0.25],
    [0.125, 0.625],
    [0.5625, 0.625],
    [0.1875, 1.0],
]))
# F5Q = sympy.Matrix([[
#     -(s**2 + s * t - 8 * s - 3 * t - 2) / 8,
#     -(25 * s**2 + 20 * s * t - t**2 - 34 * s - 28 * t - 3) / 32,
# ]])
SURFACE5Q = bezier.Surface(np.array([
    [0.25, 0.09375],
    [0.75, 0.625],
    [1.125, 0.375],
    [0.4375, 0.53125],
    [0.875, 0.75],
    [0.625, 1.0],
]))
# F6Q = sympy.Matrix([[
#     (s**2 + s * t + 3 * s + 2 * t) / 4,
#     -(3 * s**2 + 3 * s * t - 3 * s - 4 * t) / 4,
# ]])
SURFACE6Q = bezier.Surface(np.array([
    [0.0, 0.0],
    [0.375, 0.375],
    [1.0, 0.0],
    [0.25, 0.5],
    [0.75, 0.5],
    [0.5, 1.0],
]))
# F7Q = sympy.Matrix([[
#     -(s**2 + s * t + 3 * s + 2 * t - 4) / 4,
#     (8 * s**2 + 8 * s * t - 8 * s - 9 * t + 3) / 8,
# ]])
SURFACE7Q = bezier.Surface(np.array([
    [1.0, 0.375],
    [0.625, -0.125],
    [0.0, 0.375],
    [0.75, -0.1875],
    [0.25, -0.1875],
    [0.5, -0.75],
]))
# F8Q = sympy.Matrix([[
#     s * (7 * t + 1),
#     t * (7 * s + 1),
# ]])
SURFACE8Q = bezier.Surface(np.array([
    [0.0, 0.0],
    [0.5, 0.0],
    [1.0, 0.0],
    [0.0, 0.5],
    [4.0, 4.0],
    [0.0, 1.0],
]))
# F9Q = sympy.Matrix([[
#     (14 * s * t + 14 * t**2 + 2 * s - 12 * t + 3) / 2,
#     -t * (7 * s + 7 * t - 8),
# ]])
SURFACE9Q = bezier.Surface(np.array([
    [1.5, 0.0],
    [2.0, 0.0],
    [2.5, 0.0],
    [-1.5, 4.0],
    [2.5, 0.5],
    [2.5, 1.0],
]))
# F10Q = sympy.Matrix([[
#     -(2 * s + t - 2) / 2,
#     -(2 * s**2 + 2 * s*t - 2 * s + 3 * t) / 4,
# ]])
SURFACE10Q = bezier.Surface(np.array([
    [1.0, 0.0],
    [0.5, 0.25],
    [0.0, 0.0],
    [0.75, -0.375],
    [0.25, -0.375],
    [0.5, -0.75],
]))
# F11Q = sympy.Matrix([[
#     (16 * s - t**2 + 4 * t) / 16,
#     (8 * s**2 + 8 * s * t - 8 * s + 12 * t + 3) / 16,
# ]])
SURFACE11Q = bezier.Surface(np.array([
    [0.0, 0.1875],
    [0.5, -0.0625],
    [1.0, 0.1875],
    [0.125, 0.5625],
    [0.625, 0.5625],
    [0.1875, 0.9375],
]))
# F12Q = sympy.Matrix([[
#     -(2 * s + t - 2) / 2,
#     -(8 * s**2 + 8 * s * t - 8 * s + 12 * t - 1) / 16,
# ]])
SURFACE12Q = bezier.Surface(np.array([
    [1.0, 0.0625],
    [0.5, 0.3125],
    [0.0, 0.0625],
    [0.75, -0.3125],
    [0.25, -0.3125],
    [0.5, -0.6875],
]))
# F13Q = sympy.Matrix([[
#     (2 * s + t + 1) / 4,
#     (4 * s**2 + 4 * s * t + t**2 - 4 * s + 14 * t + 5) / 32,
# ]])
# NOTE: The bottom edge of this surface lies entirely on the
#       bottom edge of SURFACE4Q.
SURFACE13Q = bezier.Surface(np.array([
    [0.25, 0.15625],
    [0.5, 0.09375],
    [0.75, 0.15625],
    [0.375, 0.375],
    [0.625, 0.375],
    [0.5, 0.625],
]))
# F14Q = F13Q + sympy.Matrix([[0, 1]]) / 16
SURFACE14Q = bezier.Surface(np.array([
    [0.25, 0.21875],
    [0.5, 0.15625],
    [0.75, 0.21875],
    [0.375, 0.4375],
    [0.625, 0.4375],
    [0.5, 0.6875],
]))
# F15Q = sympy.Matrix([[
#     -2 * (s + t) * (t - 1),
#     (7 * s + 8 * t) / 4,
# ]])
SURFACE15Q = bezier.Surface(np.array([
    [0.0, 0.0],
    [1.0, 0.875],
    [2.0, 1.75],
    [1.0, 1.0],
    [1.0, 1.875],
    [0.0, 2.0],
]))
# F16Q = sympy.Matrix([[
#     2 * s**2 + 2 * s * t - 2 * s - 2 * t + 1,
#     (8 * s + 7 * t) / 4,
# ]])
SURFACE16Q = bezier.Surface(np.array([
    [1.0, 0.0],
    [0.0, 1.0],
    [1.0, 2.0],
    [0.0, 0.875],
    [0.0, 1.875],
    [-1.0, 1.75],
]))
# F17Q = sympy.Matrix([[
#     (19 * s - 19 * t + 32) / 64,
#     (2 *s * t + 5 * s + 5 * t - 6) / 8,
# ]])
SURFACE17Q = bezier.Surface(np.array([
    [0.5, -0.75],
    [0.6484375, -0.4375],
    [0.796875, -0.125],
    [0.3515625, -0.4375],
    [0.5, 0.0],
    [0.203125, -0.125],
]))
# F18Q = sympy.Matrix([[
#     -(4 * s + t - 3) / 4,
#     -(16 * s**2 + 16 * s * t - 8 * s + 27 * t - 3) / 32,
# ]])
SURFACE18Q = bezier.Surface(np.array([
    [0.75, 0.09375],
    [0.25, 0.21875],
    [-0.25, -0.15625],
    [0.625, -0.328125],
    [0.125, -0.453125],
    [0.5, -0.75],
]))
# F19Q = sympy.Matrix([[
#     (t - s) / 2,
#     -(s**2 + s * t + 2 * s + 6 * t) / 8,
# ]])
SURFACE19Q = bezier.Surface(np.array([
    [0.0, 0.0],
    [-0.25, -0.125],
    [-0.5, -0.375],
    [0.25, -0.375],
    [0.0, -0.5625],
    [0.5, -0.75],
]))
# F20Q = sympy.Matrix([[
#     -s**2 - s * t - 2 * t + 1, -s * (s + t - 2),
# ]])
SURFACE20Q = bezier.Surface(np.array([
    [1.0, 0.0],
    [1.0, 1.0],
    [0.0, 1.0],
    [0.0, 0.0],
    [-0.5, 0.5],
    [-1.0, 0.0],
]))
# F21Q = sympy.Matrix([[
#     -(3 * s**2 + 3 * s * t + 3 * t - 2) / 2,
#     -(6 * s**2 + 6 * s * t - 12 * s - t) / 4,
# ]])
SURFACE21Q = bezier.Surface(np.array([
    [1.0, 0.0],
    [1.0, 1.5],
    [-0.5, 1.5],
    [0.25, 0.125],
    [-0.5, 0.875],
    [-0.5, 0.25],
]))
# F22Q = sympy.Matrix([[
#     -5 * (2 * t - 5) * (2 * s + t - 1) / 16,
#     -5 * (16 * s**2 + 16 * s * t + 2 * t**2 - 16 * s - 37 * t + 4) / 64,
# ]])
SURFACE22Q = bezier.Surface(np.array([
    [-1.5625, -0.3125],
    [0.0, 0.3125],
    [1.5625, -0.3125],
    [-0.46875, 1.1328125],
    [0.46875, 1.1328125],
    [0.0, 2.421875],
]))
# F23Q = sympy.Matrix([[
#     (t + 2) * (2 * s + t - 1) / 2,
#     (4 * s**2 + 4 * s * t - 3 * t**2 - 4 * s + 17 * t + 1) / 8,
# ]])
SURFACE23Q = bezier.Surface(np.array([
    [-1.0, 0.125],
    [0.0, -0.125],
    [1.0, 0.125],
    [-0.75, 1.1875],
    [0.75, 1.1875],
    [0.0, 1.875],
]))
# F24Q = sympy.Matrix([[
#     -(42 * s**2 - 10 * s * t - 33 * t**2 + 16 * s + 128 * t - 96) / 128,
#     (3 * s**2 + 24 * s * t + 11 * t**2 - 18 * t + 25) / 32,
# ]])
SURFACE24Q = bezier.Surface(np.array([
    [0.75, 0.78125],
    [0.6875, 0.78125],
    [0.296875, 0.875],
    [0.25, 0.5],
    [0.2265625, 0.875],
    [0.0078125, 0.5625],
]))
# F25Q = sympy.Matrix([[
#     -(s**2 + 38 * s * t + 44 * t**2 + 40 * s + 8 * t - 62) / 64,
#     (2 * s**2 + 12 * s * t + t**2 - 24 * s - 56 * t + 62) / 64,
# ]])
SURFACE25Q = bezier.Surface(np.array([
    [0.96875, 0.96875],
    [0.65625, 0.78125],
    [0.328125, 0.625],
    [0.90625, 0.53125],
    [0.296875, 0.4375],
    [0.15625, 0.109375],
]))
# F26Q = sympy.Matrix([[s * (t + 1), t * (s + 1)]])
SURFACE26Q = bezier.Surface(np.array([
    [0.0, 0.0],
    [0.5, 0.0],
    [1.0, 0.0],
    [0.0, 0.5],
    [1.0, 1.0],
    [0.0, 1.0],
]))
# F27Q = sympy.Matrix([[s * t + t**2 - 2 * t + 2, s * t + t**2 + s]])
SURFACE27Q = bezier.Surface(np.array([
    [2.0, 0.0],
    [2.0, 0.5],
    [2.0, 1.0],
    [1.0, 0.0],
    [1.5, 1.0],
    [1.0, 1.0],
]))
# F28Q = sympy.Matrix([[
#     -(4 * s**2 + 6 * s * t - 3 * t**2 - 2 * s + 24 * t - 24) / 16,
#     (3 * s * t + 3 * t**2 + 8 * s - 3 * t) / 8,
# ]])
SURFACE28Q = bezier.Surface(np.array([
    [1.5, 0.0],
    [1.5625, 0.5],
    [1.375, 1.0],
    [0.75, -0.1875],
    [0.625, 0.5],
    [0.1875, 0.0],
]))
# F29Q = sympy.Matrix([[
#     (s**2 - 7 * s * t + 13 * s + 4 * t - 4) / 8,
#     -(7 * s * t - t**2 - 4 * s - 13 * t + 4) / 8,
# ]])
SURFACE29Q = bezier.Surface(np.array([
    [-0.5, -0.5],
    [0.3125, -0.25],
    [1.25, 0.0],
    [-0.25, 0.3125],
    [0.125, 0.125],
    [0.0, 1.25],
]))
# F30Q = sympy.Matrix([[(7 * s - 2) / 8, (7 * t - 2) / 8]])
# NOTE: This is a degree elevated linear surface.
SURFACE30Q = bezier.Surface(np.array([
    [-0.25, -0.25],
    [0.1875, -0.25],
    [0.625, -0.25],
    [-0.25, 0.1875],
    [0.1875, 0.1875],
    [-0.25, 0.625],
]))
# F31Q = sympy.Matrix([[
#     -(2 * t**2 + 2 * s * t - 8 * s - t) / 8,
#     s + t - 1,
# ]])
SURFACE31Q = bezier.Surface(np.array([
    [0.0, -1.0],
    [0.5, -0.5],
    [1.0, 0.0],
    [0.0625, -0.5],
    [0.4375, 0.0],
    [-0.125, 0.0],
]))


Intersected = collections.namedtuple(
    'Intersected',
    ['start_vals', 'end_vals', 'nodes', 'edge_pairs'])


def make_plots(surface1, surface2, intersections, failed=True):
    if not CONFIG.running:
        return

    ax = surface1.plot(64)
    surface2.plot(64, ax=ax)

    color = None
    for intersection in intersections:
        intersection.plot(64, color=color, ax=ax)
        # Color is (R,G,B,A) but we just want (R,G,B).
        color = ax.patches[-1].get_facecolor()[:3]

    plt.axis('scaled')
    _plot_helpers.add_plot_boundary(ax)

    if CONFIG.save_plot:
        CONFIG.save_fig()
    else:
        if failed:
            plt.title(CONFIG.current_test + ': failed')
        else:
            plt.title(CONFIG.current_test)
        plt.show()

    plt.close(ax.figure)


def make_curved_polygon(surface1, surface2,
                        start_vals, end_vals, edge_pairs):
    base_edges = (
        surface1._get_edges(),
        surface2._get_edges(),
    )

    info = six.moves.zip(edge_pairs, start_vals, end_vals)
    edges = []
    for edge_pair, start_val, end_val in info:
        surf_index, edge_index = edge_pair
        base_edge = base_edges[surf_index][edge_index]
        edges.append(base_edge.specialize(start_val, end_val))

    return bezier.CurvedPolygon(*edges)


def surface_surface_check(surface1, surface2,
                          start_vals, end_vals, nodes, edge_pairs):
    intersected = Intersected(start_vals, end_vals, nodes, edge_pairs)
    surface_surface_check_multi(surface1, surface2, intersected)


def surface_surface_check_multi(surface1, surface2, *all_intersected):
    # pylint: disable=too-many-locals
    assert surface1.is_valid
    assert surface2.is_valid

    intersections = surface1.intersect(surface2)
    assert len(intersections) == len(all_intersected)
    edges = (
        surface1._get_edges(),
        surface2._get_edges(),
    )

    for intersection, intersected in six.moves.zip(
            intersections, all_intersected):
        assert isinstance(intersection, bezier.CurvedPolygon)

        start_vals = intersected.start_vals
        end_vals = intersected.end_vals
        nodes = intersected.nodes
        edge_pairs = intersected.edge_pairs

        info = six.moves.zip(
            intersection._edges, edge_pairs, start_vals, end_vals, nodes)
        num_edges = len(intersection._edges)
        assert num_edges == len(edge_pairs)
        assert num_edges == len(start_vals)
        assert num_edges == len(end_vals)
        assert num_edges == len(nodes)
        for edge, edge_pair, start_val, end_val, node in info:
            surf_index, edge_index = edge_pair
            expected = edges[surf_index][edge_index]
            assert expected is edge.root

            CONFIG.assert_close(edge.start, start_val)
            CONFIG.assert_close(edge.end, end_val)
            CONFIG.assert_close(edge._nodes[0, 0], node[0])
            CONFIG.assert_close(edge._nodes[0, 1], node[1])

    make_plots(surface1, surface2, intersections, failed=False)
    # pylint: enable=too-many-locals


def test_surfaces1Q_and_3Q():
    _, s_val1 = runtime_utils.real_roots([36, -48, 4, 200, -131])
    s_val2, = runtime_utils.real_roots([49, 91, 320, -244])

    t_val1, _ = runtime_utils.real_roots([9, -18, 5, -28, 12])
    t_val2, = runtime_utils.real_roots([49, 63, 88, -128])
    start_vals = np.array([s_val2, 0.0, 0.0, t_val1])
    end_vals = np.array([1.0, 1.0, s_val1, t_val2])

    x_val1 = 0.5 * (1.0 - s_val1) * (s_val1 + 2.0)
    y_val1 = 0.5 * s_val1 * (3.0 - s_val1)
    x_val2 = 0.5 * s_val2 * (1.0 - s_val2)
    y_val2 = 1.0 - s_val2
    nodes = np.array([
        [x_val2, y_val2],
        [0.0, 0.0],
        [1.0, 0.0],
        [x_val1, y_val1],
    ])
    edge_pairs = (
        (0, 2),
        (1, 0),
        (0, 1),
        (1, 2),
    )
    # NOTE: We require a bit more wiggle room for these roots.
    with CONFIG.wiggle(45):
        surface_surface_check(SURFACE1Q, SURFACE3Q,
                              start_vals, end_vals, nodes, edge_pairs)


def test_surfaces1L_and_3L():
    start_vals = np.array([0.125, 0.25, 0.0])
    end_vals = np.array([0.875, 1.0, 0.75])

    nodes = np.array([
        [0.25, 0.75],
        [0.25, 0.0],
        [1.0, 0.0],
    ])
    edge_pairs = (
        (1, 2),
        (0, 0),
        (0, 1),
    )
    surface_surface_check(SURFACE1L, SURFACE3L,
                          start_vals, end_vals, nodes, edge_pairs)


def test_surfaces1Q_and_2Q():
    # pylint: disable=too-many-locals
    s_val1, _ = runtime_utils.real_roots([3, -21, 10])
    _, s_val2 = runtime_utils.real_roots([12, -24, 0, 140, -105])
    s_val3, _ = runtime_utils.real_roots([12, -72, 56, -100, 23])
    _, s_val4 = runtime_utils.real_roots([12, 24, -88, 156, -81])
    _, s_val5 = runtime_utils.real_roots([9, -54, 213, -12, -96])
    s_val6, _ = runtime_utils.real_roots([12, -24, 24, -140, 23])

    _, t_val1 = runtime_utils.real_roots([9, 39, -38])
    t_val2, _ = runtime_utils.real_roots([9, -18, -3, -116, 20])
    _, t_val3 = runtime_utils.real_roots([9, 30, -119, 272, -128])
    t_val4, _ = runtime_utils.real_roots([9, -66, 25, -160, 64])
    _, t_val5 = runtime_utils.real_roots([9, -66, 181, 36, -44])
    t_val6, _ = runtime_utils.real_roots([9, -18, -3, -116, 84])
    start_vals = np.array(
        [s_val6, t_val5, s_val1, t_val2, s_val3, t_val4])
    end_vals = np.array(
        [s_val5, t_val1, s_val2, t_val3, s_val4, t_val6])

    x_val1 = 0.5 * s_val6 * (1.0 - s_val6)
    x_val2 = 0.5 * s_val5 * (1.0 - s_val5)
    x_val5 = 0.5 * (1.0 - s_val3) * (s_val3 + 2.0)
    x_val6 = 0.5 * (1.0 - s_val4) * (s_val4 + 2.0)

    y_val1 = 1.0 - s_val6
    y_val2 = 1.0 - s_val5
    y_val3 = 0.5 * s_val1 * (s_val1 - 1.0)
    y_val4 = 0.5 * s_val2 * (s_val2 - 1.0)
    y_val5 = 0.5 * s_val3 * (3.0 - s_val3)
    y_val6 = 0.5 * s_val4 * (3.0 - s_val4)

    nodes = np.array([
        [x_val1, y_val1],
        [x_val2, y_val2],
        [s_val1, y_val3],
        [s_val2, y_val4],
        [x_val5, y_val5],
        [x_val6, y_val6],
    ])
    edge_pairs = (
        (0, 2),
        (1, 0),
        (0, 0),
        (1, 1),
        (0, 1),
        (1, 2),
    )
    # NOTE: We require a bit more wiggle room for these roots.
    with CONFIG.wiggle(9):
        surface_surface_check(SURFACE1Q, SURFACE2Q,
                              start_vals, end_vals, nodes, edge_pairs)
    # pylint: enable=too-many-locals


def test_surfaces10Q_and_18Q():
    start_vals = np.array([0.0, 0.25, 0.0])
    end_vals = np.array([1.0, 1.0, 1.0])

    nodes = np.array([
        [0.5, -0.75],
        [0.75, 0.09375],
        [0.0, 0.0],
    ])
    edge_pairs = (
        (1, 2),
        (0, 0),
        (0, 1),
    )
    with pytest.raises(NotImplementedError) as exc_info:
        surface_surface_check(SURFACE10Q, SURFACE18Q,
                              start_vals, end_vals, nodes, edge_pairs)

    assert str(exc_info.value).startswith(TANGENT_FAILURE)
    intersection = make_curved_polygon(
        SURFACE10Q, SURFACE18Q,
        start_vals, end_vals, edge_pairs)
    make_plots(SURFACE10Q, SURFACE18Q, [intersection])


def test_surfaces10Q_and_19Q():
    with pytest.raises(NotImplementedError) as exc_info:
        surface_surface_check_multi(SURFACE10Q, SURFACE19Q)

    assert exc_info.value.args == PARALLEL_FAILURE
    make_plots(SURFACE10Q, SURFACE19Q, [])


def test_surfaces3Q_and_4Q():
    _, s_val2 = runtime_utils.real_roots([25, -130, 321, -88, -96])
    _, s_val3 = runtime_utils.real_roots([49, 14, 145, 1008, -468])

    t_val2, _ = runtime_utils.real_roots([100, 360, 712, -2988, 169])
    _, t_val3 = runtime_utils.real_roots([49, -532, 412, 37200, -26352])
    start_vals = np.array([s_val3, t_val2, 0.0])
    end_vals = np.array([s_val2, 1.0, t_val3])

    x_val2 = 0.125 * (s_val2 - 1.0) * (5.0 * s_val2 - 8.0)
    x_val3 = 0.125 * (s_val3 - 1.0) * (5.0 * s_val3 - 8.0)
    y_val2 = 0.125 * (1.0 - s_val2) * (7.0 * s_val2 + 8.0)
    y_val3 = 0.125 * (1.0 - s_val3) * (7.0 * s_val3 + 8.0)
    nodes = np.array([
        [x_val3, y_val3],
        [x_val2, y_val2],
        [1.0, 0.25],
    ])
    edge_pairs = (
        (0, 2),
        (1, 0),
        (1, 1),
    )
    # NOTE: We require a bit more wiggle room for these roots.
    with CONFIG.wiggle(32):
        surface_surface_check(SURFACE3Q, SURFACE4Q,
                              start_vals, end_vals, nodes, edge_pairs)


def test_surfaces1Q_and_5L():
    s_val4, _ = runtime_utils.real_roots([1, -3, 1])
    t_val4, _ = runtime_utils.real_roots([1764, -3108, 1049])
    start_vals = np.array([0.5, 0.0, 0.0, t_val4])
    end_vals = np.array([1.0, 1.0, s_val4, 4.0 / 7.0])

    x_val4 = 0.5 * (1.0 - s_val4) * (s_val4 + 2.0)
    nodes = np.array([
        [0.125, 0.5],
        [0.0, 0.0],
        [1.0, 0.0],
        [x_val4, 0.5],
    ])
    edge_pairs = (
        (0, 2),
        (0, 0),
        (0, 1),
        (1, 1),
    )
    with pytest.raises(NotImplementedError) as exc_info:
        surface_surface_check(SURFACE1Q, SURFACE5L,
                              start_vals, end_vals, nodes, edge_pairs)

    assert str(exc_info.value).startswith(TANGENT_FAILURE)
    intersection = make_curved_polygon(
        SURFACE1Q, SURFACE5L,
        start_vals, end_vals, edge_pairs)
    make_plots(SURFACE1Q, SURFACE5L, [intersection])


def test_surfaces3Q_and_5Q():
    s_val3, _ = runtime_utils.real_roots([25, -130, 167, -302, 57])
    s_val4, _ = runtime_utils.real_roots([25, -130, 901, -1212, 279])

    _, t_val3 = runtime_utils.real_roots([25, -20, -1064, 7800, -6012])
    t_val4, _ = runtime_utils.real_roots([25, -2340, 58908, -105840, 11664])
    start_vals = np.array([t_val4, 0.0, 0.0, s_val3])
    end_vals = np.array([1.0, 1.0, t_val3, s_val4])

    x_val3 = 0.125 * (s_val3 - 1.0) * (5.0 * s_val3 - 8.0)
    x_val4 = 0.125 * (s_val4 - 1.0) * (5.0 * s_val4 - 8.0)
    y_val3 = 0.125 * (1.0 - s_val3) * (7.0 * s_val3 + 8.0)
    y_val4 = 0.125 * (1.0 - s_val4) * (7.0 * s_val4 + 8.0)
    nodes = np.array([
        [x_val4, y_val4],
        [0.25, 0.09375],
        [1.125, 0.375],
        [x_val3, y_val3],
    ])
    edge_pairs = (
        (1, 2),
        (1, 0),
        (1, 1),
        (0, 2),
    )
    surface_surface_check(SURFACE3Q, SURFACE5Q,
                          start_vals, end_vals, nodes, edge_pairs)


def test_surfaces1L_and_2L():
    start_vals = np.array([7.59375, 1.0, 3.0, 4.5, 0.0]) / 9.0
    end_vals = np.array([27.0, 19.0, 11.8125, 27.0, 13.5]) / 27.0

    nodes = np.array([
        [0.0, 0.46875],
        [0.0, 0.0],
        [2.0, 1.0],
        [1.6875, 1.3125],
        [0.375, 1.125],
    ]) / 3.0
    edge_pairs = (
        (0, 2),
        (1, 0),
        (0, 1),
        (1, 1),
        (1, 2),
    )
    surface_surface_check(SURFACE1L, SURFACE2L,
                          start_vals, end_vals, nodes, edge_pairs)


def test_surfaces20Q_and_21Q():
    start_vals = np.array([0.0, 0.0, 4.0 / 5.0, 0.0])
    end_vals = np.array([1.0, 0.5, 1.0, 1.0])

    nodes = np.array([
        [1.0, 0.0],
        [0.0, 1.0],
        [-0.5, 0.5],
        [-0.5, 0.25],
    ])
    edge_pairs = (
        (0, 0),
        (0, 1),
        (1, 1),
        (1, 2),
    )
    with pytest.raises(NotImplementedError) as exc_info:
        surface_surface_check(SURFACE20Q, SURFACE21Q,
                              start_vals, end_vals, nodes, edge_pairs)

    assert exc_info.value.args == PARALLEL_FAILURE
    intersection = make_curved_polygon(
        SURFACE20Q, SURFACE21Q,
        start_vals, end_vals, edge_pairs)
    make_plots(SURFACE20Q, SURFACE21Q, [intersection])


def test_surfaces4L_and_22Q():
    start_vals = np.array([0.0, 0.0, 0.0])
    end_vals = np.array([1.0, 1.0, 1.0])

    nodes = np.array([
        [-1.25, 0.0],
        [1.25, 0.0],
        [0.0, 2.1875],
    ])
    edge_pairs = (
        (0, 0),
        (0, 1),
        (0, 2),
    )
    with pytest.raises(NotImplementedError) as exc_info:
        surface_surface_check(SURFACE4L, SURFACE22Q,
                              start_vals, end_vals, nodes, edge_pairs)

    assert str(exc_info.value).startswith(TANGENT_FAILURE)
    intersection = make_curved_polygon(
        SURFACE4L, SURFACE22Q,
        start_vals, end_vals, edge_pairs)
    make_plots(SURFACE4L, SURFACE22Q, [intersection])


def test_surfaces4L_and_23Q():
    start_vals = np.array([0.0, 0.0, 0.0])
    end_vals = np.array([1.0, 1.0, 1.0])

    nodes = np.array([
        [-1.0, 0.125],
        [1.0, 0.125],
        [0.0, 1.875],
    ])
    edge_pairs = (
        (1, 0),
        (1, 1),
        (1, 2),
    )
    with pytest.raises(NotImplementedError) as exc_info:
        surface_surface_check(SURFACE4L, SURFACE23Q,
                              start_vals, end_vals, nodes, edge_pairs)

    assert str(exc_info.value).startswith(TANGENT_FAILURE)
    intersection = make_curved_polygon(
        SURFACE4L, SURFACE23Q,
        start_vals, end_vals, edge_pairs)
    make_plots(SURFACE4L, SURFACE23Q, [intersection])


def test_surfaces6Q_and_7Q():
    # pylint: disable=too-many-locals
    s_val3, _ = runtime_utils.real_roots([1, -13, 2])
    _, s_val4 = runtime_utils.real_roots([7, 5, -10])
    s_val5, s_val6 = runtime_utils.real_roots([4, 120, 1592, -1908, 489])
    s_val7, _ = runtime_utils.real_roots([64, -1232, 297])
    _, s_val8 = runtime_utils.real_roots([576, 784, -871])

    t_val3, _ = runtime_utils.real_roots([1, -102, 25])
    _, t_val4 = runtime_utils.real_roots([49, 68, -76])
    t_val6, t_val5 = runtime_utils.real_roots([4, -104, 1504, -1548, 369])
    t_val7, _ = runtime_utils.real_roots([4, -20, 3])
    _, t_val8 = runtime_utils.real_roots([12, 4, -13])

    start_vals1 = np.array([s_val8, 3.0 / 17.0, s_val3, t_val5])
    end_vals1 = np.array([14.0 / 17.0, t_val3, s_val5, t_val8])
    start_vals2 = np.array([t_val7, s_val6, t_val4, 3.0 / 17.0])
    end_vals2 = np.array([t_val6, s_val4, 14.0 / 17.0, s_val7])

    x_val3 = 0.25 * s_val3 * (3.0 + s_val3)
    x_val4 = 0.25 * s_val4 * (3.0 + s_val4)
    x_val5 = 0.25 * s_val5 * (3.0 + s_val5)
    x_val6 = 0.25 * s_val6 * (3.0 + s_val6)
    y_val3 = 0.75 * s_val3 * (1.0 - s_val3)
    y_val4 = 0.75 * s_val4 * (1.0 - s_val4)
    y_val5 = 0.75 * s_val5 * (1.0 - s_val5)
    y_val6 = 0.75 * s_val6 * (1.0 - s_val6)

    nodes1 = np.array([
        [0.5 - 0.5 * s_val8, 1.0 - s_val8],
        [3.0 / 34.0, 3.0 / 17.0],
        [x_val3, y_val3],
        [x_val5, y_val5],
    ])
    edge_pairs1 = (
        (0, 2),
        (1, 1),
        (0, 0),
        (1, 0),
    )

    nodes2 = np.array([
        [1.0 - 0.5 * s_val7, s_val7],
        [x_val6, y_val6],
        [x_val4, y_val4],
        [31.0 / 34.0, 3.0 / 17.0],
    ])
    edge_pairs2 = (
        (1, 0),
        (0, 0),
        (1, 2),
        (0, 1),
    )
    intersected1 = Intersected(start_vals1, end_vals1, nodes1, edge_pairs1)
    intersected2 = Intersected(start_vals2, end_vals2, nodes2, edge_pairs2)
    # NOTE: We require a bit more wiggle room for these roots.
    with CONFIG.wiggle(25):
        surface_surface_check_multi(SURFACE6Q, SURFACE7Q,
                                    intersected1, intersected2)
    # pylint: enable=too-many-locals


def test_surfaces8Q_and_9Q():
    s_val2, s_val3 = runtime_utils.real_roots([28, -24, 1])
    t_val3, t_val2 = runtime_utils.real_roots([28, -32, 5])
    start_vals = np.array([11.0 / 14.0, s_val2, 1.0 / 14.0, s_val3])
    end_vals = np.array([t_val2, 3.0 / 14.0, t_val3, 13.0 / 14.0])

    x_val2 = (1.0 - s_val2) * (1.0 + 7.0 * s_val2)
    x_val3 = (1.0 - s_val3) * (1.0 + 7.0 * s_val3)
    y_val2 = s_val2 * (8.0 - 7.0 * s_val2)
    y_val3 = s_val3 * (8.0 - 7.0 * s_val3)
    nodes = np.array([
        [15.0 / 28.0, 39.0 / 28.0],
        [x_val2, y_val2],
        [55.0 / 28.0, 39.0 / 28.0],
        [x_val3, y_val3],
    ])
    edge_pairs = (
        (1, 2),
        (0, 1),
        (1, 2),
        (0, 1),
    )
    surface_surface_check(SURFACE8Q, SURFACE9Q,
                          start_vals, end_vals, nodes, edge_pairs)


def test_surfaces4Q_and_10Q():
    surface_surface_check_multi(SURFACE4Q, SURFACE10Q)


def test_surfaces11Q_and_12Q():
    s_val1, s_val2 = runtime_utils.real_roots([8, -8, 1])
    start_vals = np.array([s_val1, s_val1])
    end_vals = np.array([s_val2, s_val2])

    nodes = np.array([
        [s_val2, 0.125],
        [s_val1, 0.125],
    ])
    edge_pairs = (
        (1, 0),
        (0, 0),
    )
    surface_surface_check(SURFACE11Q, SURFACE12Q,
                          start_vals, end_vals, nodes, edge_pairs)


def test_surfaces3Q_and_13Q():
    start_vals = np.array([0.0, 0.0, 0.0])
    end_vals = np.array([1.0, 1.0, 1.0])

    nodes = np.array([
        [0.25, 0.15625],
        [0.75, 0.15625],
        [0.5, 0.625],
    ])
    edge_pairs = (
        (1, 0),
        (1, 1),
        (1, 2),
    )
    surface_surface_check(SURFACE3Q, SURFACE13Q,
                          start_vals, end_vals, nodes, edge_pairs)


def test_surfaces10Q_and_17Q():
    start_vals = np.array([0.0, 0.0, 0.0])
    end_vals = np.array([1.0, 1.0, 1.0])

    nodes = np.array([
        [0.5, -0.75],
        [0.796875, -0.125],
        [0.203125, -0.125],
    ])
    edge_pairs = (
        (1, 0),
        (1, 1),
        (1, 2),
    )
    surface_surface_check(SURFACE10Q, SURFACE17Q,
                          start_vals, end_vals, nodes, edge_pairs)


def test_surfaces3Q_and_14Q():
    start_vals = np.array([0.0, 0.0, 0.0])
    end_vals = np.array([1.0, 1.0, 1.0])

    nodes = np.array([
        [0.25, 0.21875],
        [0.75, 0.21875],
        [0.5, 0.6875],
    ])
    edge_pairs = (
        (1, 0),
        (1, 1),
        (1, 2),
    )

    surface_surface_check(SURFACE3Q, SURFACE14Q,
                          start_vals, end_vals, nodes, edge_pairs)


def test_surfaces15Q_and_16Q():
    # pylint: disable=too-many-locals
    s_val4, _ = runtime_utils.real_roots([49, -120, 32])
    _, s_val5 = runtime_utils.real_roots([1, 70, -39])
    s_val6, _ = runtime_utils.real_roots([2, -18, 1])
    _, s_val7 = runtime_utils.real_roots([14, 2, -9])

    t_val4, _ = runtime_utils.real_roots([14, -30, 7])
    _, t_val5 = runtime_utils.real_roots([2, 14, -15])
    t_val6, _ = runtime_utils.real_roots([1, -72, 32])
    _, t_val7 = runtime_utils.real_roots([49, 22, -39])

    start_vals1 = np.array([0.25, t_val4, 0.5, t_val7])
    end_vals1 = np.array([s_val4, 0.5, s_val7, 0.75])
    start_vals2 = np.array([s_val6, 0.5, s_val5, 0.25])
    end_vals2 = np.array([0.5, t_val5, 0.75, t_val6])

    nodes1 = np.array([
        [0.5, 0.4375],
        [2.0 * s_val4, 1.75 * s_val4],
        [0.5, 1.0],
        [2.0 * s_val7 * (1.0 - s_val7), 2.0 - 2.0 * s_val7],
    ])
    edge_pairs1 = (
        (0, 0),
        (1, 0),
        (0, 2),
        (1, 2),
    )

    nodes2 = np.array([
        [2.0 * s_val6 * (1.0 - s_val6), 2.0 - 2.0 * s_val6],
        [0.5, 1.0],
        [2.0 - 2.0 * s_val5, 1.75 + 0.25 * s_val5],
        [0.5, 1.9375],
    ])
    edge_pairs2 = (
        (0, 2),
        (1, 0),
        (0, 1),
        (1, 1),
    )
    intersected1 = Intersected(start_vals1, end_vals1, nodes1, edge_pairs1)
    intersected2 = Intersected(start_vals2, end_vals2, nodes2, edge_pairs2)

    with pytest.raises(NotImplementedError) as exc_info:
        surface_surface_check_multi(SURFACE15Q, SURFACE16Q,
                                    intersected1, intersected2)

    assert exc_info.value.args == (BAD_TANGENT,)
    intersection1 = make_curved_polygon(
        SURFACE15Q, SURFACE16Q,
        start_vals1, end_vals1, edge_pairs1)
    intersection2 = make_curved_polygon(
        SURFACE15Q, SURFACE16Q,
        start_vals2, end_vals2, edge_pairs2)
    make_plots(SURFACE15Q, SURFACE16Q, [intersection1, intersection2])
    # pylint: enable=too-many-locals


def test_surfaces24Q_and_25Q():
    _, _, s_val1, _ = runtime_utils.real_roots(
        [81, 72, -4640, -1168, 1036])
    _, s_val2 = runtime_utils.real_roots(
        [121, -14740, 618410, -9692, -153359])
    _, t_val1, _, _ = runtime_utils.real_roots(
        [27, -1116, 12020, -10224, 2256])
    _, t_val2 = runtime_utils.real_roots(
        [11, -1232, 132116, 315936, -31348])
    start_vals = np.array([s_val2, 0.0, t_val1, 0.0])
    end_vals = np.array([1.0, s_val1, 1.0, t_val2])

    x_val1 = 0.015625 * (4.0 - 3.0 * s_val1) * (7.0 * s_val1 + 12.0)
    y_val1 = 0.03125 * (3.0 * s_val1 * s_val1 + 25.0)
    x_val2 = 0.0078125 * (33.0 * s_val2 * s_val2 + 62.0 * s_val2 + 1.0)
    y_val2 = 0.03125 * (11.0 * s_val2 * s_val2 - 4.0 * s_val2 + 18.0)
    nodes = np.array([
        [x_val2, y_val2],
        [0.75, 0.78125],
        [x_val1, y_val1],
        [0.328125, 0.625],
    ])
    edge_pairs = (
        (0, 2),
        (0, 0),
        (1, 0),
        (1, 1),
    )

    # NOTE: We require a bit more wiggle room for these roots.
    with CONFIG.wiggle(22):
        surface_surface_check(SURFACE24Q, SURFACE25Q,
                              start_vals, end_vals, nodes, edge_pairs)


def test_surfaces1L_and_6L():
    start_vals = np.array([0.75, 0.0, 0.75, 0.0])
    end_vals = np.array([1.0, 0.25, 1.0, 0.25])

    nodes = np.array([
        [0.0, 0.25],
        [0.0, 0.0],
        [0.25, 0.0],
        [0.25, 0.25],
    ])
    edge_pairs = (
        (0, 2),
        (0, 0),
        (1, 2),
        (1, 0),
    )
    surface_surface_check(SURFACE1L, SURFACE6L,
                          start_vals, end_vals, nodes, edge_pairs)


def test_surfaces26Q_and_27Q():
    surface_surface_check_multi(SURFACE26Q, SURFACE27Q)


def test_surfaces1L_and_28Q():
    _, s_val3 = runtime_utils.real_roots([5, 30, -13])
    t_val3, _ = runtime_utils.real_roots([5, -40, 22])
    start_vals = np.array([t_val3, 0.1875, 0.0])
    end_vals = np.array([1.0, 1.0, s_val3])

    nodes = np.array([
        [1.0 - s_val3, s_val3],
        [0.1875, 0.0],
        [1.0, 0.0],
    ])
    edge_pairs = (
        (1, 1),
        (0, 0),
        (0, 1),
    )
    surface_surface_check(SURFACE1L, SURFACE28Q,
                          start_vals, end_vals, nodes, edge_pairs)


def test_surfaces1L_and_29Q():
    s_val1, s_val2 = runtime_utils.real_roots([128, -128, 7])
    t_val1, t_val2 = runtime_utils.real_roots([8, -8, 1])
    start_vals = np.array([s_val2, 0.0, 0.0, 0.0, t_val1])
    end_vals = np.array([1.0, 1.0, 1.0, s_val1, t_val2])

    nodes = np.array([
        [1.0 - s_val2, s_val2],
        [0.0, 1.0],
        [0.0, 0.0],
        [1.0, 0.0],
        [1.0 - s_val1, s_val1],
    ])
    edge_pairs = (
        (0, 1),
        (0, 2),
        (0, 0),
        (0, 1),
        (1, 1),
    )
    surface_surface_check(SURFACE1L, SURFACE29Q,
                          start_vals, end_vals, nodes, edge_pairs)


def test_surfaces30Q_and_31Q():
    start_vals = np.array([13.0 / 56.0, 0.0, 5.0 / 9.0, 0.0])
    end_vals = np.array([1.0, 2.0 / 7.0, 1.0, 0.25])

    nodes = np.array([
        [-0.046875, -0.25],
        [0.625, -0.25],
        [0.375, 0.0],
        [-0.125, 0.0],
    ])
    edge_pairs = (
        (0, 0),
        (0, 1),
        (1, 1),
        (1, 2),
    )

    surface_surface_check(SURFACE30Q, SURFACE31Q,
                          start_vals, end_vals, nodes, edge_pairs)


if __name__ == '__main__':
    CONFIG.run(globals())
